# pylint: disable=missing-module-docstring
from bo4e.enum.strenum import StrEnum


class Kundengruppe(StrEnum):
    """
    Kundengruppe für eine Marktlokation (orientiert sich an den Standard-Lastprofilen).
    """

    RLM = "RLM"
    """Kunde mit registrierender Leistungsmessung (kein SLP)"""
    RLM_KOMMUNAL = "RLM_KOMMUNAL"
    """Kommunale Abnahmestelle leistungsgemessen"""
    SLP_KOMMUNAL = "SLP_KOMMUNAL"
    """Kommunale Abnahmestelle nicht leistungsgemessen"""
    SLP_S_G0 = "SLP_S_G0"
    """Gewerbe allgemein"""
    SLP_S_G1 = "SLP_S_G1"
    """Werktags"""
    SLP_S_G2 = "SLP_S_G2"
    """Verbrauch in Abendstunden"""
    SLP_S_G3 = "SLP_S_G3"
    """Gewerbe durchlaufend"""
    SLP_S_G4 = "SLP_S_G4"
    """Laden, Friseur"""
    SLP_S_G5 = "SLP_S_G5"
    """Bäckerei mit Backstube"""
    SLP_S_G6 = "SLP_S_G6"
    """Wochenendbetrieb"""
    SLP_S_G7 = "SLP_S_G7"
    """Mobilfunksendestation"""
    SLP_S_L0 = "SLP_S_L0"
    """Landwirtschaft allgemein"""
    SLP_S_L1 = "SLP_S_L1"
    """Landwirtschaft mit Milchwirtschaft/Nebenerwerbs-Tierzucht"""
    SLP_S_L2 = "SLP_S_L2"
    """Übrige Landwirtschaftsbetriebe"""
    SLP_S_H0 = "SLP_S_H0"
    """Haushalt allgemein"""
    SLP_S_SB = "SLP_S_SB"
    """Straßenbeleuchtung"""
    SLP_S_HZ = "SLP_S_HZ"
    """Nachtspeicherheizung"""
    SLP_S_WP = "SLP_S_WP"
    """Wärmepumpe"""
    SLP_S_EM = "SLP_S_EM"
    """Elektromobilität"""
    SLP_S_HZ_GEM = "SLP_S_HZ_GEM"
    """Nachtspeicherheizung gemeinsame Messung"""
    SLP_G_GKO = "SLP_G_GKO"
    """Gebietskörpersch., Kreditinst. u. Versich., Org. o. Erwerbszw. & öff. Einr."""
    SLP_G_STANDARD = "SLP_G_STANDARD"
    """Standardkundengruppe für Gas"""
    SLP_G_GHA = "SLP_G_GHA"
    """Einzelhandel, Großhandel"""
    SLP_G_GMK = "SLP_G_GMK"
    """Metall, Kfz"""
    SLP_G_GBD = "SLP_G_GBD"
    """sonst. betr. Dienstleistungen"""
    SLP_G_GGA = "SLP_G_GGA"
    """Beherbergung"""
    SLP_G_GBH = "SLP_G_GBH"
    """Gaststätten"""
    SLP_G_GBA = "SLP_G_GBA"
    """Bäckereien"""
    SLP_G_GWA = "SLP_G_GWA"
    """Wäschereien"""
    SLP_G_GGB = "SLP_G_GGB"
    """Gartenbau"""
    SLP_G_GPD = "SLP_G_GPD"
    """Papier und Druck"""
    SLP_G_GMF = "SLP_G_GMF"
    """haushaltsähnliche Gewerbebetriebe"""
    SLP_G_HEF = "SLP_G_HEF"
    """Einfamilienhaushalt"""
    SLP_G_HMF = "SLP_G_HMF"
    """Mehrfamilienhaushalt"""
    SLP_G_HKO = "SLP_G_HKO"
    """Kochgas"""



"""
Contains Steuerbetrag class
"""

from decimal import Decimal
from typing import TYPE_CHECKING, Annotated, Literal, Optional

from pydantic import Field

from ..enum.comtyp import ComTyp
from ..utils import postprocess_docstring
from .com import COM

if TYPE_CHECKING:
    from ..enum.steuerkennzeichen import Steuerkennzeichen
    from ..enum.waehrungscode import Waehrungscode


# pylint: disable=too-few-public-methods


@postprocess_docstring
class Steuerbetrag(COM):
    """
    Abbildung eines Steuerbetrages.

    .. raw:: html

        <object data="../_static/images/bo4e/com/Steuerbetrag.svg" type="image/svg+xml"></object>

    .. HINT::
        `Steuerbetrag JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/BO4E/BO4E-Schemas/{__gh_version__}/src/bo4e_schemas/com/Steuerbetrag.json>`_

    """

    typ: Annotated[Literal[ComTyp.STEUERBETRAG], Field(alias="_typ")] = ComTyp.STEUERBETRAG

    steuerkennzeichen: Optional["Steuerkennzeichen"] = None
    """Kennzeichnung des Steuersatzes, bzw. Verfahrens."""
    basiswert: Optional[Decimal] = None
    """Nettobetrag für den die Steuer berechnet wurde. Z.B. 100"""
    steuerwert: Optional[Decimal] = None
    """Aus dem Basiswert berechnete Steuer. Z.B. 19 (bei UST_19)"""
    waehrung: Optional["Waehrungscode"] = None
    """Währung. Z.B. Euro."""



from decimal import Decimal

import pytest

from bo4e import Geraet, Menge, Mengeneinheit, Tarifeinschraenkung, Voraussetzungen
from tests.serialization_helper import assert_serialization_roundtrip
from tests.test_geraet import example_geraet


class TestTarifeinschraenkung:
    @pytest.mark.parametrize(
        "tarifeinschraenkung",
        [
            pytest.param(
                Tarifeinschraenkung(
                    zusatzprodukte=["foo", "bar"],
                    voraussetzungen=[Voraussetzungen.ALTVERTRAG, Voraussetzungen.DIREKTVERTRIEB],
                    einschraenkungzaehler=[
                        example_geraet,
                        Geraet(geraetenummer="197foo"),
                    ],
                    einschraenkungleistung=[
                        Menge(wert=Decimal(12.5), einheit=Mengeneinheit.MWH),
                        Menge(wert=Decimal(30), einheit=Mengeneinheit.KWH),
                    ],
                ),
                id="maximal attributes",
            ),
        ],
    )
    def test_serialization_roundtrip(self, tarifeinschraenkung: Tarifeinschraenkung) -> None:
        """
        Test de-/serialisation of Tarifeinschraenkung
        """
        assert_serialization_roundtrip(tarifeinschraenkung)



# pylint: disable=missing-module-docstring
from bo4e.enum.strenum import StrEnum


class Messwertstatuszusatz(StrEnum):
    """
    Aufzählung von zusätzlichen Informationen zum Status, beispielsweise in Lastgängen oder Zählwerkständen.
    """

    Z84_LEERSTAND = "Z84_LEERSTAND"
    """Z84_LEERSTAND"""
    Z85_REALERZAEHLERUEBERLAUFGEPRUEFT = "Z85_REALERZAEHLERUEBERLAUFGEPRUEFT"
    """Z85_REALERZAEHLERUEBERLAUFGEPRUEFT"""
    Z86_PLAUSIBELWGKONTROLLABLESUNG = "Z86_PLAUSIBELWGKONTROLLABLESUNG"
    """Z86_PLAUSIBELWGKONTROLLABLESUNG"""
    Z87_PLAUSIBELWGKUNDENHINWEIS = "Z87_PLAUSIBELWGKUNDENHINWEIS"
    """Z87_PLAUSIBELWGKUNDENHINWEIS"""
    ZC3_AUSTAUSCHDESERSATZWERTES = "ZC3_AUSTAUSCHDESERSATZWERTES"
    """ZC3_AUSTAUSCHDESERSATZWERTES"""
    Z88_VERGLEICHSMESSUNG_GEEICHT = "Z88_VERGLEICHSMESSUNG(GEEICHT)"
    """Z88_VERGLEICHSMESSUNG(GEEICHT)"""
    Z89_VERGLEICHSMESSUNG_NICHTGEEICHT = "Z89_VERGLEICHSMESSUNG(NICHTGEEICHT)"
    """Z89_VERGLEICHSMESSUNG(NICHTGEEICHT)"""
    Z90_MESSWERTNACHBILDUNGAUSGEEICHTENWERTEN = "Z90_MESSWERTNACHBILDUNGAUSGEEICHTENWERTEN"
    """Z90_MESSWERTNACHBILDUNGAUSGEEICHTENWERTEN"""
    Z91_MESSWERTNACHBILDUNGAUSNICHTGEEICHTENWERTEN = "Z91_MESSWERTNACHBILDUNGAUSNICHTGEEICHTENWERTEN"
    """Z91_MESSWERTNACHBILDUNGAUSNICHTGEEICHTENWERTEN"""
    Z92_INTERPOLATION = "Z92_INTERPOLATION"
    """Z92_INTERPOLATION"""
    Z93_HALTEWERT = "Z93_HALTEWERT"
    """Z93_HALTEWERT"""
    Z94_BILANZIERUNGNETZABSCHNITT = "Z94_BILANZIERUNGNETZABSCHNITT"
    """Z94_BILANZIERUNGNETZABSCHNITT"""
    Z95_HISTORISCHEMESSWERTE = "Z95_HISTORISCHEMESSWERTE"
    """Z95_HISTORISCHEMESSWERTE"""
    ZJ2_STATISTISCHEMETHODE = "ZJ2_STATISTISCHEMETHODE"
    """ZJ2_STATISTISCHEMETHODE"""
    Z74_KEINZUGANG = "Z74_KEINZUGANG"
    """Z74_KEINZUGANG"""
    Z75_KOMMUNIKATIONSSTOERUNG = "Z75_KOMMUNIKATIONSSTOERUNG"
    """Z75_KOMMUNIKATIONSSTOERUNG"""
    Z76_NETZAUSFALL = "Z76_NETZAUSFALL"
    """Z76_NETZAUSFALL"""
    Z77_SPANNUNGSAUSFALL = "Z77_SPANNUNGSAUSFALL"
    """Z77_SPANNUNGSAUSFALL"""
    Z78_GERAETEWECHSEL = "Z78_GERAETEWECHSEL"
    """Z78_GERAETEWECHSEL"""
    Z79_KALIBRIERUNG = "Z79_KALIBRIERUNG"
    """Z79_KALIBRIERUNG"""
    Z80_GERAETARBEITETAUSSERHALBDERBETRIEBSBEDINGUNGEN = "Z80_GERAETARBEITETAUSSERHALBDERBETRIEBSBEDINGUNGEN"
    """Z80_GERAETARBEITETAUSSERHALBDERBETRIEBSBEDINGUNGEN"""
    Z81_MESSEINRICHTUNGGESTOERT_DEFEKT = "Z81_MESSEINRICHTUNGGESTOERT_DEFEKT"
    """Z81_MESSEINRICHTUNGGESTOERT_DEFEKT"""
    Z82_UNSICHERHEITMESSUNG = "Z82_UNSICHERHEITMESSUNG"
    """Z82_UNSICHERHEITMESSUNG"""
    Z98_BERUECKSICHTIGUNGSTOERMENGENZAEHLWERK = "Z98_BERUECKSICHTIGUNGSTOERMENGENZAEHLWERK"
    """Z98_BERUECKSICHTIGUNGSTOERMENGENZAEHLWERK"""
    Z99_MENGENUMWERTUNGUNVOLLSTAENDIG = "Z99_MENGENUMWERTUNGUNVOLLSTAENDIG"
    """Z99_MENGENUMWERTUNGUNVOLLSTAENDIG"""
    ZA0_UHRZEITGESTELLT_SYNCHRONISATION = "ZA0_UHRZEITGESTELLT_SYNCHRONISATION"
    """ZA0_UHRZEITGESTELLT_SYNCHRONISATION"""
    ZA1_MESSWERTUNPLAUSIBEL = "ZA1_MESSWERTUNPLAUSIBEL"
    """ZA1_MESSWERTUNPLAUSIBEL"""
    ZC2_TARIFSCHALTGERAETDEFEKT = "ZC2_TARIFSCHALTGERAETDEFEKT"
    """ZC2_TARIFSCHALTGERAETDEFEKT"""
    ZC4_IMPULSWERTIGKEITNICHTAUSREICHEND = "ZC4_IMPULSWERTIGKEITNICHTAUSREICHEND"
    """ZC4_IMPULSWERTIGKEITNICHTAUSREICHEND"""
    ZA3_FALSCHERWANDLERFAKTOR = "ZA3_FALSCHERWANDLERFAKTOR"
    """ZA3_FALSCHERWANDLERFAKTOR"""
    ZA4_FEHLERHAFTEABLESUNG = "ZA4_FEHLERHAFTEABLESUNG"
    """ZA4_FEHLERHAFTEABLESUNG"""
    ZA5_AENDERUNGDERBERECHNUNG = "ZA5_AENDERUNGDERBERECHNUNG"
    """ZA5_AENDERUNGDERBERECHNUNG"""
    ZA6_UMBAUDERMESSLOKATION = "ZA6_UMBAUDERMESSLOKATION"
    """ZA6_UMBAUDERMESSLOKATION"""
    ZA7_DATENBEARBEITUNGSFEHLER = "ZA7_DATENBEARBEITUNGSFEHLER"
    """ZA7_DATENBEARBEITUNGSFEHLER"""
    ZA8_BRENNWERTKORREKTUR = "ZA8_BRENNWERTKORREKTUR"
    """ZA8_BRENNWERTKORREKTUR"""
    ZA9_Z_ZAHL_KORREKTUR = "ZA9_Z-ZAHL-KORREKTUR"
    """ZA9_Z-ZAHL-KORREKTUR"""
    ZB0_STOERUNG_DEFEKTMESSEINRICHTUNG = "ZB0_STOERUNG_DEFEKTMESSEINRICHTUNG"
    """ZB0_STOERUNG_DEFEKTMESSEINRICHTUNG"""
    ZB9_AENDERUNGTARIFSCHALTZEITEN = "ZB9_AENDERUNGTARIFSCHALTZEITEN"
    """ZB9_AENDERUNGTARIFSCHALTZEITEN"""
    ZG3_UMSTELLUNGGASQUALITAET = "ZG3_UMSTELLUNGGASQUALITAET"
    """ZG3_UMSTELLUNGGASQUALITAET"""



from datetime import datetime, timezone
from decimal import Decimal

import pytest

from bo4e import (
    Betrag,
    Geschaeftspartner,
    Marktlokation,
    Menge,
    Mengeneinheit,
    Messlokation,
    NetznutzungRechnungsart,
    NetznutzungRechnungstyp,
    Preis,
    Rechnung,
    Rechnungsposition,
    Rechnungsstatus,
    Rechnungstyp,
    Sparte,
    Steuerbetrag,
    Steuerkennzeichen,
    Waehrungscode,
    Zeitraum,
)
from tests.serialization_helper import assert_serialization_roundtrip


class TestRechnung:
    @pytest.mark.parametrize(
        "rechnung",
        [
            pytest.param(
                Rechnung(
                    rechnungstitel="Hüpfburg",
                    rechnungsstatus=Rechnungsstatus.UNGEPRUEFT,
                    ist_storno=True,
                    rechnungsnummer="202201211701",
                    rechnungsdatum=datetime.today().replace(tzinfo=timezone.utc),
                    faelligkeitsdatum=datetime.today().replace(tzinfo=timezone.utc),
                    rechnungstyp=Rechnungstyp.ENDKUNDENRECHNUNG,
                    original_rechnungsnummer="RE-2022-01-21_1701",
                    rechnungsperiode=Zeitraum(dauer="P21D"),
                    rechnungsersteller=Geschaeftspartner(),
                    rechnungsempfaenger=Geschaeftspartner(),
                    gesamtnetto=Betrag(wert=Decimal(12.5), waehrung=Waehrungscode.EUR),
                    gesamtsteuer=Betrag(wert=Decimal(12.5), waehrung=Waehrungscode.EUR),
                    gesamtbrutto=Betrag(wert=Decimal(12.5), waehrung=Waehrungscode.EUR),
                    vorausgezahlt=Betrag(wert=Decimal(12.5), waehrung=Waehrungscode.EUR),
                    rabatt_brutto=Betrag(wert=Decimal(12.5), waehrung=Waehrungscode.EUR),
                    zu_zahlen=Betrag(wert=Decimal(12.5), waehrung=Waehrungscode.EUR),
                    steuerbetraege=[
                        Steuerbetrag(
                            steuerkennzeichen=Steuerkennzeichen.UST_19,
                            basiswert=Decimal(20.25),
                            waehrung=Waehrungscode.EUR,
                            steuerwert=Decimal(10.5),
                        )
                    ],
                    rechnungspositionen=[
                        Rechnungsposition(
                            positionsnummer=1,
                            lieferung_von=datetime(2021, 3, 15, tzinfo=timezone.utc),
                            lieferung_bis=datetime(2022, 3, 15, tzinfo=timezone.utc),
                            positionstext="Besonders wertvolle Rechnungsposition",
                            positions_menge=Menge(),
                            einzelpreis=Preis(),
                            teilsumme_netto=Betrag(),
                            teilsumme_steuer=Steuerbetrag(),
                        )
                    ],
                    sparte=Sparte.STROM,
                    netznutzungrechnungsart=NetznutzungRechnungsart.SELBSTAUSGESTELLT,
                    netznutzungrechnungstyp=NetznutzungRechnungstyp.TURNUSRECHNUNG,
                    ist_original=True,
                    ist_simuliert=True,
                    marktlokation=Marktlokation(
                        marktlokations_id="51238696781",
                    ),
                    messlokation=Messlokation(
                        messlokations_id="DE00056266802AO6G56M11SN51G21M24S",
                    ),
                ),
                id="maximal attributes",
            ),
            pytest.param(
                Rechnung(
                    ist_storno=True,
                    rechnungsnummer="202201211701",
                    rechnungsdatum=datetime.today().replace(tzinfo=timezone.utc),
                    faelligkeitsdatum=datetime.today().replace(tzinfo=timezone.utc),
                    rechnungstyp=Rechnungstyp.ENDKUNDENRECHNUNG,
                    original_rechnungsnummer="RE-2022-01-21_1701",
                    rechnungsperiode=Zeitraum(dauer="P21D"),
                    rechnungsersteller=Geschaeftspartner(),
                    rechnungsempfaenger=Geschaeftspartner(),
                    gesamtnetto=Betrag(wert=Decimal(12.5), waehrung=Waehrungscode.EUR),
                    gesamtsteuer=Betrag(wert=Decimal(12.5), waehrung=Waehrungscode.EUR),
                    gesamtbrutto=Betrag(wert=Decimal(12.5), waehrung=Waehrungscode.EUR),
                    zu_zahlen=Betrag(wert=Decimal(12.5), waehrung=Waehrungscode.EUR),
                    rechnungspositionen=[
                        Rechnungsposition(
                            positionsnummer=1,
                            lieferung_von=datetime(2021, 3, 15, tzinfo=timezone.utc),
                            lieferung_bis=datetime(2022, 3, 15, tzinfo=timezone.utc),
                            positionstext="Besonders wertvolle Rechnungsposition",
                            positions_menge=Menge(),
                            einzelpreis=Preis(),
                            teilsumme_netto=Betrag(),
                            teilsumme_steuer=Steuerbetrag(),
                        )
                    ],
                    sparte=Sparte.STROM,
                    netznutzungrechnungsart=NetznutzungRechnungsart.SELBSTAUSGESTELLT,
                    netznutzungrechnungstyp=NetznutzungRechnungstyp.TURNUSRECHNUNG,
                    ist_original=True,
                    ist_simuliert=True,
                    marktlokation=Marktlokation(
                        marktlokations_id="51238696781",
                    ),
                    messlokation=Messlokation(
                        messlokations_id="DE00056266802AO6G56M11SN51G21M24S",
                    ),
                ),
                id="minimal attributes",
            ),
        ],
    )
    def test_serialization_roundtrip(self, rechnung: Rechnung) -> None:
        """
        Test de-/serialisation
        """
        assert_serialization_roundtrip(rechnung)



"""
Contains Ausschreibung class
"""

# pylint: disable=too-few-public-methods, too-many-instance-attributes
# pylint: disable=no-name-in-module
from typing import TYPE_CHECKING, Annotated, Literal, Optional

import pydantic
from pydantic import Field

from ..enum.botyp import BoTyp
from ..utils import postprocess_docstring
from .geschaeftsobjekt import Geschaeftsobjekt

if TYPE_CHECKING:
    from ..com.ausschreibungslos import Ausschreibungslos
    from ..com.zeitraum import Zeitraum
    from ..enum.ausschreibungsportal import Ausschreibungsportal
    from ..enum.ausschreibungsstatus import Ausschreibungsstatus
    from ..enum.ausschreibungstyp import Ausschreibungstyp
    from .geschaeftspartner import Geschaeftspartner


@postprocess_docstring
class Ausschreibung(Geschaeftsobjekt):
    """
    Das BO Ausschreibung dient zur detaillierten Darstellung von ausgeschriebenen Energiemengen in der Energiewirtschaft

    .. raw:: html

        <object data="../_static/images/bo4e/bo/Ausschreibung.svg" type="image/svg+xml"></object>

    .. HINT::
        `Ausschreibung JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/BO4E/BO4E-Schemas/{__gh_version__}/src/bo4e_schemas/bo/Ausschreibung.json>`_

    """

    typ: Annotated[Literal[BoTyp.AUSSCHREIBUNG], Field(alias="_typ")] = BoTyp.AUSSCHREIBUNG
    ausschreibungsnummer: Optional[str] = None
    """Vom Herausgeber der Ausschreibung vergebene eindeutige Nummer"""
    ausschreibungstyp: Optional["Ausschreibungstyp"] = None
    """Aufzählung für die Typisierung von Ausschreibungen"""
    ausschreibungsstatus: Optional["Ausschreibungsstatus"] = None
    """Bezeichnungen für die Ausschreibungsphasen"""
    ist_kostenpflichtig: Optional[bool] = None
    """Kennzeichen, ob die Ausschreibung kostenpflichtig ist"""
    veroeffentlichungszeitpunkt: Optional[pydantic.AwareDatetime] = None
    """Gibt den Veröffentlichungszeitpunkt der Ausschreibung an"""
    ausschreibender: Optional["Geschaeftspartner"] = None
    """
    Mit diesem Objekt können Geschäftspartner übertragen werden.
    Sowohl Unternehmen, als auch Privatpersonen können Geschäftspartner sein
    """
    abgabefrist: Optional["Zeitraum"] = None
    """
    Diese Komponente wird zur Abbildung von Zeiträumen in Form von Dauern oder der Angabe von Start und Ende verwendet.
    Es muss daher entweder eine Dauer oder ein Zeitraum in Form von Start und Ende angegeben sein
    """
    bindefrist: Optional["Zeitraum"] = None
    """
    Diese Komponente wird zur Abbildung von Zeiträumen in Form von Dauern oder der Angabe von Start und Ende verwendet.
    Es muss daher entweder eine Dauer oder ein Zeitraum in Form von Start und Ende angegeben sein
    """
    lose: Optional[list["Ausschreibungslos"]] = None
    """Die einzelnen Lose, aus denen sich die Ausschreibung zusammensetzt"""

    ausschreibungportal: Optional["Ausschreibungsportal"] = None
    """Aufzählung der unterstützten Ausschreibungsportale"""
    webseite: Optional[str] = None
    """Internetseite, auf der die Ausschreibung veröffentlicht wurde (falls vorhanden)"""



# pylint: disable=missing-module-docstring
from bo4e.enum.strenum import StrEnum


class TechnischeRessourceNutzung(StrEnum):
    """
    Nutzung der technischen Ressource
    """

    STROMVERBRAUCHSART = "STROMVERBRAUCHSART"
    STROMERZEUGUNGSART = "STROMERZEUGUNGSART"
    SPEICHER = "SPEICHER"



"""
BO4E-Python - Python implementation of the BO4E standard

BO4E-Python is a Python implementation of the BO4E standard. BO4E is a standard for the exchange of
business objects in the energy industry. BO4E-Python is an open-source software released under the Apache-2.0
license.

The BO4E version can be queried using `bo4e.__version__`.
"""

__all__ = [
    "Abwicklungsmodell",
    "Angebot",
    "Ausschreibung",
    "Bilanzierung",
    "Buendelvertrag",
    "Energiemenge",
    "Fremdkosten",
    "Geraet",
    "Geschaeftsobjekt",
    "Geschaeftspartner",
    "Kosten",
    "Lastgang",
    "Lokationszuordnung",
    "Marktlokation",
    "Marktteilnehmer",
    "Messlokation",
    "Netzlokation",
    "Person",
    "Preisblatt",
    "PreisblattDienstleistung",
    "PreisblattHardware",
    "PreisblattKonzessionsabgabe",
    "PreisblattMessung",
    "PreisblattNetznutzung",
    "Rechnung",
    "Region",
    "Regionaltarif",
    "Standorteigenschaften",
    "SteuerbareRessource",
    "Tarif",
    "Tarifinfo",
    "Tarifkosten",
    "Tarifpreisblatt",
    "TechnischeRessource",
    "Vertrag",
    "Zaehler",
    "Zeitreihe",
    "Adresse",
    "Angebotsposition",
    "Angebotsteil",
    "Angebotsvariante",
    "AufAbschlag",
    "AufAbschlagProOrt",
    "AufAbschlagRegional",
    "AufAbschlagstaffelProOrt",
    "Ausschreibungsdetail",
    "Ausschreibungslos",
    "Betrag",
    "COM",
    "Dienstleistung",
    "Energieherkunft",
    "Energiemix",
    "Fremdkostenblock",
    "Fremdkostenposition",
    "Geokoordinaten",
    "Katasteradresse",
    "Kontaktweg",
    "Konzessionsabgabe",
    "Kostenblock",
    "Kostenposition",
    "KriteriumWert",
    "Lastprofil",
    "MarktgebietInfo",
    "Menge",
    "PositionsAufAbschlag",
    "Preis",
    "Preisgarantie",
    "Preisposition",
    "Preisstaffel",
    "Rechnungsposition",
    "RegionaleGueltigkeit",
    "RegionalePreisgarantie",
    "RegionalePreisstaffel",
    "RegionalerAufAbschlag",
    "RegionaleTarifpreisposition",
    "Regionskriterium",
    "Sigmoidparameter",
    "StandorteigenschaftenGas",
    "StandorteigenschaftenStrom",
    "Steuerbetrag",
    "Tagesparameter",
    "Tarifberechnungsparameter",
    "Tarifeinschraenkung",
    "Tarifpreis",
    "Tarifpreisposition",
    "TarifpreispositionProOrt",
    "TarifpreisstaffelProOrt",
    "Unterschrift",
    "Verbrauch",
    "Vertragskonditionen",
    "Vertragsteil",
    "VerwendungszweckProMarktrolle",
    "Zaehlwerk",
    "Zaehlzeitregister",
    "Zeitraum",
    "Zeitreihenwert",
    "Zustaendigkeit",
    "AbgabeArt",
    "Aggregationsverantwortung",
    "Angebotsstatus",
    "Anrede",
    "ArithmetischeOperation",
    "ArtikelId",
    "AufAbschlagstyp",
    "AufAbschlagsziel",
    "Ausschreibungsportal",
    "Ausschreibungsstatus",
    "Ausschreibungstyp",
    "BDEWArtikelnummer",
    "Befestigungsart",
    "Bemessungsgroesse",
    "Bilanzierungsmethode",
    "Dienstleistungstyp",
    "EMobilitaetsart",
    "Energierichtung",
    "Erzeugungsart",
    "Fallgruppenzuordnung",
    "Gasqualitaet",
    "Gebiettyp",
    "Geraeteklasse",
    "Geraetetyp",
    "Geschaeftspartnerrolle",
    "Gueltigkeitstyp",
    "Kalkulationsmethode",
    "Konfigurationsprodukt",
    "Kontaktart",
    "Kostenklasse",
    "Kundengruppe",
    "KundengruppeKA",
    "Kundentyp",
    "Landescode",
    "Leistungstyp",
    "Lokationstyp",
    "Marktrolle",
    "Medium",
    "Mengeneinheit",
    "Mengenoperator",
    "Messart",
    "Messgroesse",
    "Messpreistyp",
    "Messwerterfassung",
    "Messwertstatus",
    "Messwertstatuszusatz",
    "Netzebene",
    "NetznutzungRechnungsart",
    "NetznutzungRechnungstyp",
    "Oekolabel",
    "Oekozertifikat",
    "Organisationstyp",
    "Preisgarantietyp",
    "Preismodell",
    "Preisstatus",
    "Preistyp",
    "Profilart",
    "Profiltyp",
    "Profilverfahren",
    "Prognosegrundlage",
    "Rechnungslegung",
    "Rechnungsstatus",
    "Rechnungstyp",
    "Regionskriteriumtyp",
    "Registeranzahl",
    "Rollencodetyp",
    "Sparte",
    "Speicherart",
    "SteuerkanalLeistungsbeschreibung",
    "Steuerkennzeichen",
    "StrEnum",
    "Tarifkalkulationsmethode",
    "Tarifmerkmal",
    "Tarifregionskriterium",
    "Tariftyp",
    "Tarifzeit",
    "TechnischeRessourceNutzung",
    "TechnischeRessourceVerbrauchsart",
    "Themengebiet",
    "Titel",
    "BoTyp",
    "ComTyp",
    "Verbrauchsart",
    "Vertragsart",
    "Vertragsform",
    "Vertragsstatus",
    "Verwendungszweck",
    "Voraussetzungen",
    "WahlrechtPrognosegrundlage",
    "Waehrungscode",
    "Waehrungseinheit",
    "Waermenutzung",
    "Zaehlerauspraegung",
    "Zaehlergroesse",
    "Zaehlertyp",
    "ZaehlertypSpezifikation",
    "Zeitreihentyp",
    "ZusatzAttribut",
    "__version__",
    "__gh_version__",
]

from pydantic import BaseModel as _PydanticBaseModel

# Import BOs
from .bo.angebot import Angebot
from .bo.ausschreibung import Ausschreibung
from .bo.bilanzierung import Bilanzierung
from .bo.buendelvertrag import Buendelvertrag
from .bo.energiemenge import Energiemenge
from .bo.fremdkosten import Fremdkosten
from .bo.geraet import Geraet
from .bo.geschaeftsobjekt import Geschaeftsobjekt
from .bo.geschaeftspartner import Geschaeftspartner
from .bo.kosten import Kosten
from .bo.lastgang import Lastgang
from .bo.lokationszuordnung import Lokationszuordnung
from .bo.marktlokation import Marktlokation
from .bo.marktteilnehmer import Marktteilnehmer
from .bo.messlokation import Messlokation
from .bo.netzlokation import Netzlokation
from .bo.person import Person
from .bo.preisblatt import Preisblatt
from .bo.preisblattdienstleistung import PreisblattDienstleistung
from .bo.preisblatthardware import PreisblattHardware
from .bo.preisblattkonzessionsabgabe import PreisblattKonzessionsabgabe
from .bo.preisblattmessung import PreisblattMessung
from .bo.preisblattnetznutzung import PreisblattNetznutzung
from .bo.rechnung import Rechnung
from .bo.region import Region
from .bo.regionaltarif import Regionaltarif
from .bo.standorteigenschaften import Standorteigenschaften
from .bo.steuerbareressource import SteuerbareRessource
from .bo.tarif import Tarif
from .bo.tarifinfo import Tarifinfo
from .bo.tarifkosten import Tarifkosten
from .bo.tarifpreisblatt import Tarifpreisblatt
from .bo.technischeressource import TechnischeRessource
from .bo.vertrag import Vertrag
from .bo.zaehler import Zaehler
from .bo.zeitreihe import Zeitreihe

# Import COMs
from .com.adresse import Adresse
from .com.angebotsposition import Angebotsposition
from .com.angebotsteil import Angebotsteil
from .com.angebotsvariante import Angebotsvariante
from .com.aufabschlag import AufAbschlag
from .com.aufabschlagproort import AufAbschlagProOrt
from .com.aufabschlagregional import AufAbschlagRegional
from .com.aufabschlagstaffelproort import AufAbschlagstaffelProOrt
from .com.ausschreibungsdetail import Ausschreibungsdetail
from .com.ausschreibungslos import Ausschreibungslos
from .com.betrag import Betrag
from .com.com import COM
from .com.dienstleistung import Dienstleistung
from .com.energieherkunft import Energieherkunft
from .com.energiemix import Energiemix
from .com.fremdkostenblock import Fremdkostenblock
from .com.fremdkostenposition import Fremdkostenposition
from .com.geokoordinaten import Geokoordinaten
from .com.katasteradresse import Katasteradresse
from .com.konfigurationsprodukt import Konfigurationsprodukt
from .com.kontaktweg import Kontaktweg
from .com.konzessionsabgabe import Konzessionsabgabe
from .com.kostenblock import Kostenblock
from .com.kostenposition import Kostenposition
from .com.kriteriumwert import KriteriumWert
from .com.lastprofil import Lastprofil
from .com.marktgebietinfo import MarktgebietInfo
from .com.menge import Menge
from .com.positionsaufabschlag import PositionsAufAbschlag
from .com.preis import Preis
from .com.preisgarantie import Preisgarantie
from .com.preisposition import Preisposition
from .com.preisstaffel import Preisstaffel
from .com.rechnungsposition import Rechnungsposition
from .com.regionalegueltigkeit import RegionaleGueltigkeit
from .com.regionalepreisgarantie import RegionalePreisgarantie
from .com.regionalepreisstaffel import RegionalePreisstaffel
from .com.regionaleraufabschlag import RegionalerAufAbschlag
from .com.regionaletarifpreisposition import RegionaleTarifpreisposition
from .com.regionskriterium import Regionskriterium
from .com.sigmoidparameter import Sigmoidparameter
from .com.standorteigenschaftengas import StandorteigenschaftenGas
from .com.standorteigenschaftenstrom import StandorteigenschaftenStrom
from .com.steuerbetrag import Steuerbetrag
from .com.tagesparameter import Tagesparameter
from .com.tarifberechnungsparameter import Tarifberechnungsparameter
from .com.tarifeinschraenkung import Tarifeinschraenkung
from .com.tarifpreis import Tarifpreis
from .com.tarifpreisposition import Tarifpreisposition
from .com.tarifpreispositionproort import TarifpreispositionProOrt
from .com.tarifpreisstaffelproort import TarifpreisstaffelProOrt
from .com.unterschrift import Unterschrift
from .com.verbrauch import Verbrauch
from .com.vertragskonditionen import Vertragskonditionen
from .com.vertragsteil import Vertragsteil
from .com.verwendungszweckpromarktrolle import VerwendungszweckProMarktrolle
from .com.zaehlwerk import Zaehlwerk
from .com.zaehlzeitregister import Zaehlzeitregister
from .com.zeitraum import Zeitraum
from .com.zeitreihenwert import Zeitreihenwert
from .com.zustaendigkeit import Zustaendigkeit

# Import Enums
from .enum.abgabeart import AbgabeArt
from .enum.abwicklungsmodell import Abwicklungsmodell
from .enum.aggregationsverantwortung import Aggregationsverantwortung
from .enum.angebotsstatus import Angebotsstatus
from .enum.anrede import Anrede
from .enum.arithmetische_operation import ArithmetischeOperation
from .enum.artikelid import ArtikelId
from .enum.aufabschlagstyp import AufAbschlagstyp
from .enum.aufabschlagsziel import AufAbschlagsziel
from .enum.ausschreibungsportal import Ausschreibungsportal
from .enum.ausschreibungsstatus import Ausschreibungsstatus
from .enum.ausschreibungstyp import Ausschreibungstyp
from .enum.bdewartikelnummer import BDEWArtikelnummer
from .enum.befestigungsart import Befestigungsart
from .enum.bemessungsgroesse import Bemessungsgroesse
from .enum.bilanzierungsmethode import Bilanzierungsmethode
from .enum.botyp import BoTyp
from .enum.comtyp import ComTyp
from .enum.dienstleistungstyp import Dienstleistungstyp
from .enum.emobilitaetsart import EMobilitaetsart
from .enum.energierichtung import Energierichtung
from .enum.erzeugungsart import Erzeugungsart
from .enum.fallgruppenzuordnung import Fallgruppenzuordnung
from .enum.gasqualitaet import Gasqualitaet
from .enum.gebiettyp import Gebiettyp
from .enum.geraeteklasse import Geraeteklasse
from .enum.geraetetyp import Geraetetyp
from .enum.geschaeftspartnerrolle import Geschaeftspartnerrolle
from .enum.gueltigkeitstyp import Gueltigkeitstyp
from .enum.kalkulationsmethode import Kalkulationsmethode
from .enum.kontaktart import Kontaktart
from .enum.kostenklasse import Kostenklasse
from .enum.kundengruppe import Kundengruppe
from .enum.kundengruppeka import KundengruppeKA
from .enum.kundentyp import Kundentyp
from .enum.landescode import Landescode
from .enum.leistungstyp import Leistungstyp
from .enum.lokationstyp import Lokationstyp
from .enum.marktrolle import Marktrolle
from .enum.medium import Medium
from .enum.mengeneinheit import Mengeneinheit
from .enum.mengenoperator import Mengenoperator
from .enum.messart import Messart
from .enum.messgroesse import Messgroesse
from .enum.messpreistyp import Messpreistyp
from .enum.messwerterfassung import Messwerterfassung
from .enum.messwertstatus import Messwertstatus
from .enum.messwertstatuszusatz import Messwertstatuszusatz
from .enum.netzebene import Netzebene
from .enum.netznutzungrechnungsart import NetznutzungRechnungsart
from .enum.netznutzungrechnungstyp import NetznutzungRechnungstyp
from .enum.oekolabel import Oekolabel
from .enum.oekozertifikat import Oekozertifikat
from .enum.organisationstyp import Organisationstyp
from .enum.preisgarantietyp import Preisgarantietyp
from .enum.preismodell import Preismodell
from .enum.preisstatus import Preisstatus
from .enum.preistyp import Preistyp
from .enum.profilart import Profilart
from .enum.profiltyp import Profiltyp
from .enum.profilverfahren import Profilverfahren
from .enum.prognosegrundlage import Prognosegrundlage
from .enum.rechnungslegung import Rechnungslegung
from .enum.rechnungsstatus import Rechnungsstatus
from .enum.rechnungstyp import Rechnungstyp
from .enum.regionskriteriumtyp import Regionskriteriumtyp
from .enum.registeranzahl import Registeranzahl
from .enum.rollencodetyp import Rollencodetyp
from .enum.sparte import Sparte
from .enum.speicherart import Speicherart
from .enum.steuerkanalleistungsbeschreibung import SteuerkanalLeistungsbeschreibung
from .enum.steuerkennzeichen import Steuerkennzeichen
from .enum.strenum import StrEnum
from .enum.tarifkalkulationsmethode import Tarifkalkulationsmethode
from .enum.tarifmerkmal import Tarifmerkmal
from .enum.tarifregionskriterium import Tarifregionskriterium
from .enum.tariftyp import Tariftyp
from .enum.tarifzeit import Tarifzeit
from .enum.technischeressourcenutzung import TechnischeRessourceNutzung
from .enum.technischeressourceverbrauchsart import TechnischeRessourceVerbrauchsart
from .enum.themengebiet import Themengebiet
from .enum.titel import Titel
from .enum.verbrauchsart import Verbrauchsart
from .enum.vertragsart import Vertragsart
from .enum.vertragsform import Vertragsform
from .enum.vertragsstatus import Vertragsstatus
from .enum.verwendungszweck import Verwendungszweck
from .enum.voraussetzungen import Voraussetzungen
from .enum.waehrungscode import Waehrungscode
from .enum.waehrungseinheit import Waehrungseinheit
from .enum.waermenutzung import Waermenutzung
from .enum.wahlrechtprognosegrundlage import WahlrechtPrognosegrundlage
from .enum.zaehlerauspraegung import Zaehlerauspraegung
from .enum.zaehlergroesse import Zaehlergroesse
from .enum.zaehlertyp import Zaehlertyp
from .enum.zaehlertypspezifikation import ZaehlertypSpezifikation
from .enum.zeitreihentyp import Zeitreihentyp
from .version import __gh_version__, __version__
from .zusatzattribut import ZusatzAttribut

# Resolve all ForwardReferences. This design prevents circular import errors.
for cls_name in __all__:
    cls = globals().get(cls_name, None)
    if cls is None or not isinstance(cls, type) or not issubclass(cls, _PydanticBaseModel):
        continue
    cls.model_rebuild(force=True)



# pylint: disable=missing-module-docstring
from bo4e.enum.strenum import StrEnum


class Messwertstatus(StrEnum):
    """
    Der Status eines Zählerstandes
    """

    ABGELESEN = "ABGELESEN"
    """ABGELESEN"""
    ERSATZWERT = "ERSATZWERT"
    """ERSATZWERT"""
    ANGABE_FUER_LIEFERSCHEIN = "ANGABE_FUER_LIEFERSCHEIN"
    """ANGABE_FUER_LIEFERSCHEIN"""
    VORSCHLAGSWERT = "VORSCHLAGSWERT"
    """VORSCHLAGSWERT"""
    NICHT_VERWENDBAR = "NICHT_VERWENDBAR"
    """NICHT_VERWENDBAR"""
    PROGNOSEWERT = "PROGNOSEWERT"
    """PROGNOSEWERT"""
    VORLAEUFIGERWERT = "VORLAEUFIGERWERT"
    """VORLAEUFIGERWERT"""
    ENERGIEMENGESUMMIERT = "ENERGIEMENGESUMMIERT"
    """ENERGIEMENGESUMMIERT"""
    FEHLT = "FEHLT"
    """FEHLT"""



# pylint:disable=missing-module-docstring

from bo4e.enum.strenum import StrEnum


class Zaehlertyp(StrEnum):
    """
    Bei diesem Enum handelt es sich um die Abbildung von Zählertypen der Sparten Strom und Gas.
    """

    DREHSTROMZAEHLER = "DREHSTROMZAEHLER"
    """Drehstromzähler - dieser Zählertyp wird für dreiphasige Ferraris-Zähler verwendet"""
    BALGENGASZAEHLER = "BALGENGASZAEHLER"
    """Balgengaszähler"""
    DREHKOLBENZAEHLER = "DREHKOLBENZAEHLER"
    """Drehkolbengaszähler"""
    LEISTUNGSZAEHLER = "LEISTUNGSZAEHLER"
    """leistungsmessender Zähler"""
    MAXIMUMZAEHLER = "MAXIMUMZAEHLER"
    """Maximumzähler"""
    TURBINENRADGASZAEHLER = "TURBINENRADGASZAEHLER"
    """Turbinenradgaszähler"""
    ULTRASCHALLGASZAEHLER = "ULTRASCHALLGASZAEHLER"
    """Ultraschallgaszähler"""
    WECHSELSTROMZAEHLER = "WECHSELSTROMZAEHLER"
    """Wechselstromzähler - dieser Zählertyp wird für einphasige Ferraris-Zähler verwendet"""
    MODERNE_MESSEINRICHTUNG = "MODERNE_MESSEINRICHTUNG"
    """Moderne Messeinrichtung"""
    INTELLIGENTES_MESSSYSTEM = "INTELLIGENTES_MESSSYSTEM"
    """Intelligentes Messsystem"""
    ELEKTRONISCHER_ZAEHLER = "ELEKTRONISCHER_ZAEHLER"
    """Elektronischer Zähler"""
    WIRBELGASZAEHLER = "WIRBELGASZAEHLER"
    """Wirbelgaszähler"""
    WASSERZAEHLER = "WASSERZAEHLER"
    """Wasserzähler"""



# pylint: disable=missing-module-docstring
from bo4e.enum.strenum import StrEnum


class ArithmetischeOperation(StrEnum):
    """
    Mit dieser Aufzählung können arithmetische Operationen festgelegt werden.
    """

    ADDITION = "ADDITION"
    """Es wird addiert"""
    SUBTRAKTION = "SUBTRAKTION"
    """Es wird subtrahiert"""
    MULTIPLIKATION = "MULTIPLIKATION"
    """Es wird multipliziert"""
    DIVISION = "DIVISION"
    """Es wird dividiert"""



# pylint: disable=missing-module-docstring
from bo4e.enum.strenum import StrEnum


class Steuerkennzeichen(StrEnum):
    """
    Zur Kennzeichnung verschiedener Steuersätze und Verfahren.
    """

    UST_0 = "UST_0"
    """Keine Umsatzsteuer, bzw. nicht steuerbar."""
    UST_19 = "UST_19"
    """Umsatzsteuer 19%"""
    UST_16 = "UST_16"
    """Umsatzsteuer 16%"""
    UST_7 = "UST_7"
    """Umsatzsteuer 7%"""
    VST_0 = "VST_0"
    """Keine Vorsteuer, bzw. nicht steuerbar."""
    VST_19 = "VST_19"
    """Vorsteuer 19%"""
    VST_16 = "VST_16"
    """Vorsteuer 16%"""
    VST_7 = "VST_7"
    """Vorsteuer 7%"""
    RCV = "RCV"
    """Reverse Charge Verfahren (Umkehrung der Steuerpflicht)"""



"""
Contains class Bilanzierung
"""

from datetime import datetime
from decimal import Decimal
from typing import Annotated, Optional

from pydantic import Field

from bo4e.bo.geschaeftsobjekt import Geschaeftsobjekt
from bo4e.com.lastprofil import Lastprofil
from bo4e.com.menge import Menge
from bo4e.enum.aggregationsverantwortung import Aggregationsverantwortung
from bo4e.enum.fallgruppenzuordnung import Fallgruppenzuordnung
from bo4e.enum.profiltyp import Profiltyp
from bo4e.enum.prognosegrundlage import Prognosegrundlage
from bo4e.enum.wahlrechtprognosegrundlage import WahlrechtPrognosegrundlage
from bo4e.enum.zeitreihentyp import Zeitreihentyp

from ..enum.abwicklungsmodell import Abwicklungsmodell
from ..enum.botyp import BoTyp


class Bilanzierung(Geschaeftsobjekt):
    """
    Das BO Bilanzierung erfasst alle relevanten Informationen zur Bilanzierung.

    .. raw:: html

        <object data="../_static/images/bo4e/bo/Bilanzierung.svg" type="image/svg+xml"></object>

    .. HINT::
        `Lastprofil JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-Schemas/{__gh_version__}/src/bo4e_schemas/bo/Bilanzierung.json>`_

    """

    typ: Annotated[Optional[BoTyp], Field(alias="_typ")] = BoTyp.BILANZIERUNG

    marktlokations_id: Optional[str] = None  #:  ID der Marktlokation
    lastprofil: Optional[list[Lastprofil]] = None  #: Eine Liste der verwendeten Lastprofile (SLP, SLP/TLP, ALP etc.)
    bilanzierungsbeginn: Optional[datetime] = None  #: Inklusiver Start der Bilanzierung
    bilanzierungsende: Optional[datetime] = None  #: Exklusives Ende der Bilanzierung
    bilanzkreis: Optional[str] = None  #: Bilanzkreis
    jahresverbrauchsprognose: Optional[Menge] = None  #: Jahresverbrauchsprognose
    temperatur_arbeit: Optional[Menge] = None  #: Temperatur Arbeit
    kundenwert: Optional[Menge] = None  #: Kundenwert
    verbrauchsaufteilung: Optional[Decimal] = None
    """
    Verbrauchsaufteilung in % zwischen SLP und TLP-Profil.

    1. [Gemessene Energiemenge der OBIS "nicht Schwachlast"] * [Verbrauchsaufteilung in % / 100%]
    = [zu verlagernde Energiemenge]
    2. [Gemessene Energiemenge der OBIS "Schwachlast"] - [zu verlagernde Energiemenge]
    = [Ermittelte Energiemenge für Schwachlast]
    3. [Gemessene Energiemenge der OBIS "nicht Schwachlast"] + [zu verlagernde Energiemenge]
    = [Ermittelte Energiemenge für nicht Schwachlast]
    """
    zeitreihentyp: Optional[Zeitreihentyp] = None  #: Zeitreihentyp (SLS, TLS, etc.)
    aggregationsverantwortung: Optional[Aggregationsverantwortung] = None  #: Aggregationsverantwortung
    prognosegrundlage: Optional[Prognosegrundlage] = None  #: Prognosegrundlage
    details_prognosegrundlage: Optional[list[Profiltyp]] = None
    """
    Prognosegrundlage.

    Besteht der Bedarf ein tagesparameteräbhängiges Lastprofil mit gemeinsamer Messung anzugeben,
    so ist dies über die 2 -malige Wiederholung des CAV Segments mit der Angabe der Codes E02 und E14 möglich.
    """
    wahlrecht_prognosegrundlage: Optional[WahlrechtPrognosegrundlage] = None
    """
    Wahlrecht der Prognosegrundlage.
    """
    fallgruppenzuordnung: Optional[Fallgruppenzuordnung] = None  #: Fallgruppenzuordnung (für gas RLM)
    prioritaet: Optional[int] = None  #: Priorität des Bilanzkreises (für Gas)
    grund_wahlrecht_prognosegrundlage: Optional[WahlrechtPrognosegrundlage] = None
    """
    Grund Wahlrecht der Prognosegrundlage.

    true=Wahlrecht beim Lieferanten vorhanden
    """
    abwicklungsmodell: Optional[Abwicklungsmodell] = None  #: Abwicklungsmodell



import pytest

from bo4e import Angebotsposition, Angebotsteil, Betrag, Marktlokation, Menge, Zeitraum
from tests.serialization_helper import assert_serialization_roundtrip


class TestAngebotsteil:
    @pytest.mark.parametrize(
        "angebotsteil",
        [
            pytest.param(
                Angebotsteil(
                    positionen=[Angebotsposition()],
                    anfrage_subreferenz="teststring",
                    lieferstellenangebotsteil=[Marktlokation()],
                    gesamtmengeangebotsteil=Menge(),
                    gesamtkostenangebotsteil=Betrag(),
                    lieferzeitraum=Zeitraum(),
                ),
                id="all attributes at first level",
            ),
        ],
    )
    def test_serialization_roundtrip(self, angebotsteil: Angebotsteil) -> None:
        """
        Test de-/serialisation of Angebotsteil with minimal attributes.
        """
        assert_serialization_roundtrip(angebotsteil)



from pathlib import Path

from docs.uml import PlantUMLNetwork, build_network


def test_network_build() -> None:
    """
    This test is only for debugging purposes.
    """
    project_root_dir = Path(__file__).parent.parent
    module_dir = project_root_dir / "src/bo4e"
    _network, _namespaces_to_parse = build_network(module_dir, PlantUMLNetwork)



"""
Contains Katasteradresse class
"""

from typing import Annotated, Literal, Optional

from pydantic import Field

from ..enum.comtyp import ComTyp
from ..utils import postprocess_docstring
from .com import COM

# pylint: disable=too-few-public-methods


@postprocess_docstring
class Katasteradresse(COM):
    """
    Dient der Adressierung über die Liegenschafts-Information.

    .. raw:: html

        <object data="../_static/images/bo4e/com/Katasteradresse.svg" type="image/svg+xml"></object>

    .. HINT::
        `Katasteradresse JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/BO4E/BO4E-Schemas/{__gh_version__}/src/bo4e_schemas/com/Katasteradresse.json>`_

    """

    typ: Annotated[Literal[ComTyp.KATASTERADRESSE], Field(alias="_typ")] = ComTyp.KATASTERADRESSE

    gemarkung_flur: Optional[str] = None
    flurstueck: Optional[str] = None



"""
Contains Tarifeinschraenkung
"""

from typing import TYPE_CHECKING, Annotated, Literal, Optional

from pydantic import Field

from ..enum.comtyp import ComTyp
from ..utils import postprocess_docstring
from .com import COM

if TYPE_CHECKING:
    from ..bo.geraet import Geraet
    from ..enum.voraussetzungen import Voraussetzungen
    from .menge import Menge

# pylint: disable=too-few-public-methods


@postprocess_docstring
class Tarifeinschraenkung(COM):
    """
    Mit dieser Komponente werden Einschränkungen für die Anwendung von Tarifen modelliert.

    .. raw:: html

        <object data="../_static/images/bo4e/com/Tarifeinschraenkung.svg" type="image/svg+xml"></object>

    .. HINT::
        `Tarifeinschraenkung JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/BO4E/BO4E-Schemas/{__gh_version__}/src/bo4e_schemas/com/Tarifeinschraenkung.json>`_

    """

    typ: Annotated[Literal[ComTyp.TARIFEINSCHRAENKUNG], Field(alias="_typ")] = ComTyp.TARIFEINSCHRAENKUNG

    zusatzprodukte: Optional[list[str]] = None
    """Weitere Produkte, die gemeinsam mit diesem Tarif bestellt werden können"""
    voraussetzungen: Optional[list["Voraussetzungen"]] = None
    """Voraussetzungen, die erfüllt sein müssen, damit dieser Tarif zur Anwendung kommen kann"""
    einschraenkungzaehler: Optional[list["Geraet"]] = None
    """ Liste der Zähler/Geräte, die erforderlich sind, damit dieser Tarif zur Anwendung gelangen kann.
    (Falls keine Zähler angegeben sind, ist der Tarif nicht an das Vorhandensein bestimmter Zähler gebunden.) """
    einschraenkungleistung: Optional[list["Menge"]] = None
    """ Die vereinbarte Leistung, die (näherungsweise) abgenommen wird.
    Insbesondere Gastarife können daran gebunden sein, dass die Leistung einer vereinbarten Höhe entspricht. """



"""
Contains Buendelvertrag class
"""

# pylint: disable=too-few-public-methods
# pylint: disable=no-name-in-module
from typing import TYPE_CHECKING, Annotated, Literal, Optional

import pydantic
from pydantic import Field

from ..enum.botyp import BoTyp
from ..utils import postprocess_docstring
from .geschaeftsobjekt import Geschaeftsobjekt

if TYPE_CHECKING:
    from ..com.unterschrift import Unterschrift
    from ..com.vertragskonditionen import Vertragskonditionen
    from ..enum.sparte import Sparte
    from ..enum.vertragsart import Vertragsart
    from ..enum.vertragsstatus import Vertragsstatus
    from .geschaeftspartner import Geschaeftspartner
    from .vertrag import Vertrag


@postprocess_docstring
class Buendelvertrag(Geschaeftsobjekt):
    """
    Abbildung eines Bündelvertrags.
    Es handelt sich hierbei um eine Liste von Einzelverträgen, die in einem Vertragsobjekt gebündelt sind.

    .. raw:: html

        <object data="../_static/images/bo4e/bo/Buendelvertrag.svg" type="image/svg+xml"></object>

    .. HINT::
        `Buendelvertrag JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/BO4E/BO4E-Schemas/{__gh_version__}/src/bo4e_schemas/bo/Buendelvertrag.json>`_

    """

    typ: Annotated[Literal[BoTyp.BUENDELVERTRAG], Field(alias="_typ")] = BoTyp.BUENDELVERTRAG

    # pylint: disable=duplicate-code
    vertragsnummer: Optional[str] = None
    """Eine im Verwendungskontext eindeutige Nummer für den Vertrag"""
    vertragsart: Optional["Vertragsart"] = None
    """Hier ist festgelegt, um welche Art von Vertrag es sich handelt. Z.B. Netznutzungvertrag"""
    vertragsstatus: Optional["Vertragsstatus"] = None
    """Gibt den Status des Vertrages an"""
    sparte: Optional["Sparte"] = None
    """Unterscheidungsmöglichkeiten für die Sparte"""
    vertragsbeginn: Optional[pydantic.AwareDatetime] = None
    """Gibt an, wann der Vertrag beginnt (inklusiv)"""
    vertragsende: Optional[pydantic.AwareDatetime] = None
    """Gibt an, wann der Vertrag (voraussichtlich) endet oder beendet wurde (exklusiv)"""
    vertragspartner1: Optional["Geschaeftspartner"] = None
    """
    Der "erstgenannte" Vertragspartner. In der Regel der Aussteller des Vertrags.
    Beispiel: "Vertrag zwischen Vertagspartner 1 ..."
    """
    vertragspartner2: Optional["Geschaeftspartner"] = None
    """
    Der "zweitgenannte" Vertragspartner. In der Regel der Empfänger des Vertrags.
    Beispiel "Vertrag zwischen Vertagspartner 1 und Vertragspartner 2"
    """

    einzelvertraege: Optional[list["Vertrag"]] = None
    """Die Liste mit den Einzelverträgen zu den Abnahmestellen"""
    vertragskonditionen: Optional[list["Vertragskonditionen"]] = None
    """Festlegungen zu Laufzeiten und Kündigungsfristen"""
    unterzeichnervp1: Optional[list["Unterschrift"]] = None
    """Unterzeichner des Vertragspartners1"""
    unterzeichnervp2: Optional[list["Unterschrift"]] = None
    """Unterzeichner des Vertragspartners2"""
    beschreibung: Optional[str] = None
    """Beschreibung zum Vertrag"""



"""
Contains Enums for Wahlrechtsprognosgrundlage.
"""

from bo4e.enum.strenum import StrEnum


class WahlrechtPrognosegrundlage(StrEnum):
    """
    Wahlrecht der Prognosegrundlage der Marktlokation.
    """

    DURCH_LF = "DURCH_LF"  #: Wahlrecht durch LF gegeben, remark: SG10 CAV
    DURCH_LF_NICHT_GEGEBEN = "DURCH_LF_NICHT_GEGEBEN"  #: Wahlrecht durch LF nicht gegeben, remark: CAV + ZE2
    NICHT_WEGEN_GROSSEN_VERBRAUCHS = "NICHT_WEGEN_GROSSEN_VERBRAUCHS"  # : kein WR, Verbrauch>10k, CAV+ Z55
    NICHT_WEGEN_EIGENVERBRAUCH = "NICHT_WEGEN_EIGENVERBRAUCH"  #: kein WR, Eigenverbrauch, CAV + ZC1
    NICHT_WEGEN_TAGES_VERBRAUCH = "NICHT_WEGEN_TAGES_VERBRAUCH"  # : kein WR, tagesparam.abh. Verbrauch, CAV + ZD2
    NICHT_WEGEN_ENWG = "NICHT_WEGEN_ENWG"  #: WR nicht wegen $14a EnWG, CAV + ZE3



"""
Contains Person class
"""

from datetime import date
from typing import TYPE_CHECKING, Annotated, Literal, Optional

from pydantic import Field

from ..enum.botyp import BoTyp
from ..utils import postprocess_docstring
from .geschaeftsobjekt import Geschaeftsobjekt

if TYPE_CHECKING:
    from ..com.adresse import Adresse
    from ..com.kontaktweg import Kontaktweg
    from ..com.zustaendigkeit import Zustaendigkeit
    from ..enum.anrede import Anrede
    from ..enum.titel import Titel


# pylint: disable=too-many-instance-attributes, too-few-public-methods, disable=duplicate-code


@postprocess_docstring
class Person(Geschaeftsobjekt):
    """
    Object containing information about a Person

    .. raw:: html

        <object data="../_static/images/bo4e/bo/Person.svg" type="image/svg+xml"></object>

    .. HINT::
        `Person JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/BO4E/BO4E-Schemas/{__gh_version__}/src/bo4e_schemas/bo/Person.json>`_

    """

    typ: Annotated[Literal[BoTyp.PERSON], Field(alias="_typ")] = BoTyp.PERSON
    anrede: Optional["Anrede"] = None
    """Mögliche Anrede der Person"""
    individuelle_anrede: Optional[str] = None
    """
    Im Falle einer nicht standardisierten Anrede kann hier eine frei definierbare Anrede vorgegeben werden.
    Beispiel: "Vereinsgemeinschaft", "Pfarrer", "Hochwürdigster Herr Abt".
    """
    titel: Optional["Titel"] = None
    """Möglicher Titel der Person"""
    vorname: Optional[str] = None
    """Vorname der Person"""
    nachname: Optional[str] = None
    """Nachname (Familienname) der Person"""
    kontaktwege: Optional[list["Kontaktweg"]] = None
    """Kontaktwege der Person"""
    geburtsdatum: Optional[date] = None
    """Geburtsdatum der Person"""
    kommentar: Optional[str] = None
    """Weitere Informationen zur Person"""
    zustaendigkeiten: Optional[list["Zustaendigkeit"]] = None
    """Liste der Abteilungen und Zuständigkeiten der Person"""
    adresse: Optional["Adresse"] = None
    """Adresse der Person, falls diese von der Adresse des Geschäftspartners abweicht"""



# pylint: disable=missing-module-docstring
from bo4e.enum.strenum import StrEnum


class Abwicklungsmodell(StrEnum):
    """
    Art des Abwicklungsmodell (E-Mob)
    """

    MODELL_1 = "MODELL_1"  #: Modell 1 "Bilanzierung an der Marktlokation"
    MODELL_2 = "MODELL_2"  #: Modell 2 "Bilanzierung im Bilanzierungsgebiet (BG) des LPB



"""
Contains Unterschrift class
"""

from typing import Annotated, Literal, Optional

import pydantic
from pydantic import Field

from ..enum.comtyp import ComTyp
from ..utils import postprocess_docstring
from .com import COM

# pylint: disable=too-few-public-methods


@postprocess_docstring
class Unterschrift(COM):
    """
    Modellierung einer Unterschrift, z.B. für Verträge, Angebote etc.

    .. raw:: html

        <object data="../_static/images/bo4e/com/Unterschrift.svg" type="image/svg+xml"></object>

    .. HINT::
        `Unterschrift JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/BO4E/BO4E-Schemas/{__gh_version__}/src/bo4e_schemas/com/Unterschrift.json>`_

    """

    typ: Annotated[Literal[ComTyp.UNTERSCHRIFT], Field(alias="_typ")] = ComTyp.UNTERSCHRIFT

    name: Optional[str] = None
    """Name des Unterschreibers"""

    ort: Optional[str] = None
    """Ort, an dem die Unterschrift geleistet wird"""
    datum: Optional[pydantic.AwareDatetime] = None
    """Datum der Unterschrift"""



"""
Contains Tarifinfo class
"""

# pylint: disable=too-many-instance-attributes, too-few-public-methods
# pylint: disable=no-name-in-module
from typing import TYPE_CHECKING, Annotated, Literal, Optional

import pydantic
from pydantic import Field

from ..enum.botyp import BoTyp
from ..utils import postprocess_docstring
from .geschaeftsobjekt import Geschaeftsobjekt

if TYPE_CHECKING:
    from ..com.energiemix import Energiemix
    from ..com.vertragskonditionen import Vertragskonditionen
    from ..com.zeitraum import Zeitraum
    from ..enum.kundentyp import Kundentyp
    from ..enum.registeranzahl import Registeranzahl
    from ..enum.sparte import Sparte
    from ..enum.tarifmerkmal import Tarifmerkmal
    from ..enum.tariftyp import Tariftyp
    from .marktteilnehmer import Marktteilnehmer


@postprocess_docstring
class Tarifinfo(Geschaeftsobjekt):
    """
    Das BO Tarifinfo liefert die Merkmale, die einen Endkundentarif identifizierbar machen.
    Dieses BO dient als Basis für weitere BOs mit erweiterten Anwendungsmöglichkeiten.

    .. raw:: html

        <object data="../_static/images/bo4e/bo/Tarifinfo.svg" type="image/svg+xml"></object>

    .. HINT::
        `Tarifinfo JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/BO4E/BO4E-Schemas/{__gh_version__}/src/bo4e_schemas/bo/Tarifinfo.json>`_

    """

    typ: Annotated[Literal[BoTyp.TARIFINFO], Field(alias="_typ")] = BoTyp.TARIFINFO
    bezeichnung: Optional[str] = None
    """Name des Tarifs"""
    anbietername: Optional[str] = None
    """Der Name des Marktpartners, der den Tarif anbietet"""
    sparte: Optional["Sparte"] = None
    """Strom oder Gas, etc."""
    kundentypen: Optional[list["Kundentyp"]] = None
    """Kundentypen für den der Tarif gilt, z.B. Privatkunden"""
    registeranzahl: Optional["Registeranzahl"] = None
    """Die Art des Tarifes, z.B. Eintarif oder Mehrtarif"""
    tariftyp: Optional["Tariftyp"] = None
    """Hinweis auf den Tariftyp, z.B. Grundversorgung oder Sondertarif"""
    tarifmerkmale: Optional[list["Tarifmerkmal"]] = None
    """Weitere Merkmale des Tarifs, z.B. Festpreis oder Vorkasse"""
    anbieter: Optional["Marktteilnehmer"] = None
    """Der Marktteilnehmer (Lieferant), der diesen Tarif anbietet"""

    website: Optional[str] = None
    """Internetseite auf dem der Tarif zu finden ist"""
    bemerkung: Optional[str] = None
    """Freitext"""

    zeitliche_gueltigkeit: Optional["Zeitraum"] = None
    """Angabe, in welchem Zeitraum der Tarif gültig ist"""
    energiemix: Optional["Energiemix"] = None
    """Der Energiemix, der für diesen Tarif gilt"""
    vertragskonditionen: Optional["Vertragskonditionen"] = None
    """Mindestlaufzeiten und Kündigungsfristen zusammengefasst"""
    anwendung_von: Optional[pydantic.AwareDatetime] = None
    """
    Angabe des inklusiven Zeitpunkts, ab dem der Tarif bzw. der Preis angewendet und abgerechnet wird,
    z.B. "2021-07-20T18:31:48Z"
    """



# pylint:disable=missing-module-docstring

from bo4e.enum.strenum import StrEnum


class Netzebene(StrEnum):
    """
    Auflistung möglicher Netzebenen innerhalb der Energiearten Strom und Gas.
    """

    NSP = "NSP"
    """Niederspannung; Strom"""
    MSP = "MSP"
    """Mittelspannung; Strom"""
    HSP = "HSP"
    """Hochspannung; Strom"""
    HSS = "HSS"
    """Hoechstspannung; Strom"""
    MSP_NSP_UMSP = "MSP_NSP_UMSP"
    """MS/NS Umspannung; Strom"""
    HSP_MSP_UMSP = "HSP_MSP_UMSP"
    """HS/MS Umspannung; Strom"""
    HSS_HSP_UMSP = "HSS_HSP_UMSP"
    """HOES/HS Umspannung; Strom"""
    HD = "HD"
    """Hochdruck; Gas"""
    MD = "MD"
    """Mitteldruck; Gas"""
    ND = "ND"
    """Niederdruck; Gas"""



# pylint: disable=missing-module-docstring
from bo4e.enum.strenum import StrEnum


class Preisstatus(StrEnum):
    """
    Statusinformation für Preise
    """

    VORLAEUFIG = "VORLAEUFIG"
    """vorläufig"""
    ENDGUELTIG = "ENDGUELTIG"
    """endgültig"""



from decimal import Decimal

import pytest

from bo4e import Energieherkunft, Energiemix, Oekolabel, Oekozertifikat, Sparte
from tests.serialization_helper import assert_serialization_roundtrip


class TestEnergiemix:
    @pytest.mark.parametrize(
        "energiemix",
        [
            pytest.param(
                Energiemix(
                    energiemixnummer=2,
                    energieart=Sparte.STROM,
                    bezeichnung="foo",
                    gueltigkeitsjahr=2021,
                    anteil=[
                        Energieherkunft(),
                        Energieherkunft(),
                    ],
                    oekolabel=[
                        Oekolabel.GASGREEN,
                        Oekolabel.GRUENER_STROM_GOLD,
                    ],
                    bemerkung="bar",
                    co2_emission=Decimal(40),
                    atommuell=Decimal(5),
                    website="foobar.de",
                    oekozertifikate=[Oekozertifikat.FRAUNHOFER, Oekozertifikat.FREIBERG],
                    ist_in_oeko_top_ten=True,
                ),
                id="all attributes at first level",
            ),
        ],
    )
    def test_energiemix_serialization_roundtrip(self, energiemix: Energiemix) -> None:
        """
        Test de-/serialisation of Energiehermix with minimal attributes.
        """
        assert_serialization_roundtrip(energiemix)



"""
Contains Netzlokation class
"""

from typing import TYPE_CHECKING, Annotated, Literal, Optional

from pydantic import Field

from ..enum.botyp import BoTyp
from ..utils import postprocess_docstring
from .geschaeftsobjekt import Geschaeftsobjekt

if TYPE_CHECKING:
    from ..bo.lokationszuordnung import Lokationszuordnung
    from ..com.konfigurationsprodukt import Konfigurationsprodukt
    from ..com.menge import Menge
    from ..com.verwendungszweckpromarktrolle import VerwendungszweckProMarktrolle
    from ..enum.marktrolle import Marktrolle
    from ..enum.sparte import Sparte


# pylint: disable=too-many-instance-attributes, too-few-public-methods


@postprocess_docstring
class Netzlokation(Geschaeftsobjekt):
    """
    Object containing information about a Netzlokation

    .. raw:: html

        <object data="../_static/images/bo4e/bo/Netzlokation.svg" type="image/svg+xml"></object>

    .. HINT::
        `Netzlokation JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/BO4E/BO4E-Schemas/{__gh_version__}/src/bo4e_schemas/bo/Netzlokation.json>`_

    """

    typ: Annotated[Literal[BoTyp.NETZLOKATION], Field(alias="_typ")] = BoTyp.NETZLOKATION

    netzlokations_id: Optional[str] = None
    """Identifikationsnummer einer Netzlokation, an der Energie entweder verbraucht, oder erzeugt wird"""
    sparte: Optional["Sparte"] = None
    """Sparte der Netzlokation, z.B. Gas oder Strom."""
    netzanschlussleistung: Optional["Menge"] = None
    """Netzanschlussleistungsmenge der Netzlokation"""
    grundzustaendiger_msb_codenr: Optional[str] = None
    """Codenummer des grundzuständigen Messstellenbetreibers, der für diese Netzlokation zuständig ist."""
    steuerkanal: Optional[bool] = None
    """Ob ein Steuerkanal der Netzlokation zugeordnet ist und somit die Netzlokation gesteuert werden kann."""
    obiskennzahl: Optional[str] = None
    """Die OBIS-Kennzahl für die Netzlokation"""
    verwendungszweck: Optional["VerwendungszweckProMarktrolle"] = None
    """Verwendungungszweck der Werte Netzlokation"""
    konfigurationsprodukte: Optional[list["Konfigurationsprodukt"]] = None
    """Produkt-Daten der Netzlokation"""
    eigenschaft_msb_lokation: Optional["Marktrolle"] = None
    """Eigenschaft des Messstellenbetreibers an der Lokation"""
    lokationszuordnungen: Optional[list["Lokationszuordnung"]] = None
    """Lokationszuordnung, um bspw. die zugehörigen Messlokationen anzugeben"""
    lokationsbuendel_objektcode: Optional[str] = None
    """Lokationsbuendel Code, der die Funktion dieses BOs an der Lokationsbuendelstruktur beschreibt."""



"""
A module that simplifies assertions for json serialization
"""

from __future__ import annotations

from typing import Any, Dict, Optional, TypeVar

from dictdiffer import diff  # type:ignore[import-not-found]
from pydantic import BaseModel

from bo4e.bo.geschaeftsobjekt import Geschaeftsobjekt
from bo4e.com.com import COM

T = TypeVar("T", bound=BaseModel)


def assert_serialization_roundtrip(serializable_object: T, expected_json_dict: Optional[Dict[str, Any]] = None) -> T:
    """
    Serializes the serializable_object using the provided schema,
    then asserts, that the result is equal to the expected_json_dict (if given),
    then deserializes the dictionary again and asserts the equality with the original serializable_object
    :returns the deserialized_object
    """
    json_string = serializable_object.model_dump_json(by_alias=True)
    assert json_string is not None
    actual_json_dict = serializable_object.model_dump(by_alias=True)
    assert actual_json_dict is not None

    def _remove_version_recursive_iter(value: dict[str, Any] | list[Any]) -> None:
        if isinstance(value, dict):
            value.pop("_version", None)
            for v in value.values():
                _remove_version_recursive_iter(v)
        elif isinstance(value, list):
            for v in value:
                _remove_version_recursive_iter(v)

    _remove_version_recursive_iter(actual_json_dict)

    # Check typ property for bo and com classes
    # the default value must be the corresponding enum value of the enum classes BoTyp or ComTyp
    if isinstance(serializable_object, (Geschaeftsobjekt, COM)):
        assert hasattr(
            serializable_object, "typ"
        ), f"Missing 'typ' attribute in {serializable_object.__class__.__name__}"
        assert serializable_object.typ is not None

        assert serializable_object.typ.value.lower() == serializable_object.__class__.__name__.lower(), (
            f"typ value '{serializable_object.typ.value}' does not match class name "
            f"'{serializable_object.__class__.__name__}'"
        )

    if expected_json_dict is not None:
        assert all([(k in json_string) for k in expected_json_dict.keys()])
        assert actual_json_dict == expected_json_dict, (
            f"actual_json_dict != expected_json_dict\n\t" f"diff = {list(diff(expected_json_dict, actual_json_dict))}"
        )
        # this (diff(...)) contains the difference between two dicts -> just for easier debugging

    deserialized_object = type(serializable_object).model_validate_json(json_string)
    assert isinstance(deserialized_object, type(serializable_object))
    assert deserialized_object == serializable_object
    return deserialized_object



"""
Contains Kostenposition
"""

from typing import TYPE_CHECKING, Annotated, Literal, Optional

import pydantic
from pydantic import Field

from ..enum.comtyp import ComTyp
from ..utils import postprocess_docstring
from .com import COM

if TYPE_CHECKING:

    from .betrag import Betrag
    from .menge import Menge
    from .preis import Preis

# pylint: disable=too-few-public-methods, too-many-instance-attributes


@postprocess_docstring
class Kostenposition(COM):
    """
    Diese Komponente wird zur Übertagung der Details zu einer Kostenposition verwendet.

    .. raw:: html

        <object data="../_static/images/bo4e/com/Kostenposition.svg" type="image/svg+xml"></object>

    .. HINT::
        `Kostenposition JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/BO4E/BO4E-Schemas/{__gh_version__}/src/bo4e_schemas/com/Kostenposition.json>`_

    """

    typ: Annotated[Literal[ComTyp.KOSTENPOSITION], Field(alias="_typ")] = ComTyp.KOSTENPOSITION

    positionstitel: Optional[str] = None
    """Ein Titel für die Zeile. Hier kann z.B. der Netzbetreiber eingetragen werden, wenn es sich um Netzkosten handelt."""

    betrag_kostenposition: Optional["Betrag"] = None
    """Der errechnete Gesamtbetrag der Position als Ergebnis der Berechnung <Menge * Einzelpreis> oder
    <Einzelpreis / (Anzahl Tage Jahr) * zeitmenge>"""
    # todo: validate above calculation, see https://github.com/Hochfrequenz/BO4E-python/issues/282

    artikelbezeichnung: Optional[str] = None
    """Bezeichnung für den Artikel für den die Kosten ermittelt wurden. Beispiel: Arbeitspreis HT"""

    einzelpreis: Optional["Preis"] = None
    """Der Preis für eine Einheit. Beispiele: 5,8200 ct/kWh oder 55 €/Jahr."""

    von: Optional[pydantic.AwareDatetime] = None
    """inklusiver von-Zeitpunkt der Kostenzeitscheibe"""
    bis: Optional[pydantic.AwareDatetime] = None
    """exklusiver bis-Zeitpunkt der Kostenzeitscheibe"""

    menge: Optional["Menge"] = None
    """Die Menge, die in die Kostenberechnung eingeflossen ist. Beispiel: 3.660 kWh"""

    zeitmenge: Optional["Menge"] = None
    """
    Wenn es einen zeitbasierten Preis gibt (z.B. €/Jahr), dann ist hier die Menge angegeben mit der die Kosten berechnet
    wurden. Z.B. 138 Tage.
    """

    artikeldetail: Optional[str] = None
    """Detaillierung des Artikels (optional). Beispiel: 'Drehstromzähler'"""



"""
Contains Regionaltarif class
"""

from typing import TYPE_CHECKING, Annotated, Literal, Optional

import pydantic
from pydantic import Field

from ..enum.botyp import BoTyp
from .tarifinfo import Tarifinfo

if TYPE_CHECKING:
    from ..com.regionalepreisgarantie import RegionalePreisgarantie
    from ..com.regionaleraufabschlag import RegionalerAufAbschlag
    from ..com.regionaletarifpreisposition import RegionaleTarifpreisposition
    from ..com.tarifberechnungsparameter import Tarifberechnungsparameter
    from ..com.tarifeinschraenkung import Tarifeinschraenkung


# pylint: disable=too-few-public-methods, empty-docstring
# pylint: disable=no-name-in-module


class Regionaltarif(Tarifinfo):
    """
    Abbildung eines Tarifs mit regionaler Zuordnung von Preisen und Auf- und Abschlägen.

    .. raw:: html

        <object data="../_static/images/bo4e/bo/Regionaltarif.svg" type="image/svg+xml"></object>

    .. HINT::
        `Regionaltarif JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/BO4E/BO4E-Schemas/{__gh_version__}/src/bo4e_schemas/bo/Regionaltarif.json>`_

    """

    typ: Annotated[Literal[BoTyp.REGIONALTARIF], Field(alias="_typ")] = BoTyp.REGIONALTARIF  # type: ignore[assignment]
    preisstand: Optional[pydantic.AwareDatetime] = None
    """Gibt an, wann der Preis zuletzt angepasst wurde"""
    berechnungsparameter: Optional["Tarifberechnungsparameter"] = None
    """Für die Berechnung der Kosten sind die hier abgebildeten Parameter heranzuziehen"""
    tarifpreise: Optional[list["RegionaleTarifpreisposition"]] = None
    """Die festgelegten Preise mit regionaler Eingrenzung, z.B. für Arbeitspreis, Grundpreis etc."""

    tarif_auf_abschlaege: Optional[list["RegionalerAufAbschlag"]] = None
    """Auf- und Abschläge auf die Preise oder Kosten mit regionaler Eingrenzung"""
    preisgarantien: Optional[list["RegionalePreisgarantie"]] = None
    """Festlegung von Garantien für bestimmte Preisanteile"""
    tarifeinschraenkung: Optional["Tarifeinschraenkung"] = None
    """Die Bedingungen und Einschränkungen unter denen ein Tarif angewendet werden kann"""



# pylint: disable=missing-module-docstring
from bo4e.enum.strenum import StrEnum


class Medium(StrEnum):
    """
    Gibt ein physikalisches Medium an.
    """

    STROM = "STROM"
    """STROM"""
    GAS = "GAS"
    """GAS"""
    WASSER = "WASSER"
    """WASSER"""
    DAMPF = "DAMPF"
    """DAMPF"""



# pylint:disable=missing-module-docstring

from bo4e.enum.strenum import StrEnum


class Rollencodetyp(StrEnum):
    """
    Gibt den Codetyp einer Rolle, beispielsweise einer Marktrolle, an.
    """

    BDEW = "BDEW"
    """Bundesverband der Energie- u. Wasserwirtschaft"""
    DVGW = "DVGW"
    """Deutscher Verein des Gas- und Wasserfaches"""
    GLN = "GLN"
    """Global Location Number"""



"""
Contains Enums for Profilverfahren.
"""

from bo4e.enum.strenum import StrEnum


class Profilverfahren(StrEnum):
    """
    Profilverfahren: synthetisch/ analytisch.
    """

    SYNTHETISCH = "SYNTHETISCH"  #: SLP
    ANALYTISCH = "ANALYTISCH"  #: ALP



"""
Contains Preisstaffel
"""

from decimal import Decimal
from typing import TYPE_CHECKING, Annotated, Literal, Optional

from pydantic import Field

from ..enum.comtyp import ComTyp
from ..utils import postprocess_docstring
from .com import COM

if TYPE_CHECKING:

    from .sigmoidparameter import Sigmoidparameter

# pylint: disable=too-few-public-methods


@postprocess_docstring
class Preisstaffel(COM):
    """
    Gibt die Staffelgrenzen der jeweiligen Preise an

    .. raw:: html

        <object data="../_static/images/bo4e/com/Preisstaffel.svg" type="image/svg+xml"></object>

    .. HINT::
        `Preisstaffel JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/BO4E/BO4E-Schemas/{__gh_version__}/src/bo4e_schemas/com/Preisstaffel.json>`_

    """

    typ: Annotated[Literal[ComTyp.PREISSTAFFEL], Field(alias="_typ")] = ComTyp.PREISSTAFFEL

    einheitspreis: Optional[Decimal] = None
    """Preis pro abgerechneter Mengeneinheit"""
    staffelgrenze_von: Optional[Decimal] = None
    """Inklusiver unterer Wert, ab dem die Staffel gilt"""
    staffelgrenze_bis: Optional[Decimal] = None
    """Exklusiver oberer Wert, bis zu dem die Staffel gilt"""

    sigmoidparameter: Optional["Sigmoidparameter"] = None
    """Parameter zur Berechnung des Preises anhand der Jahresmenge und weiterer netzbezogener Parameter"""

    artikel_id: Optional[str] = None
    """Standardisierte vom BDEW herausgegebene Liste, welche im Strommarkt die BDEW-Artikelnummer ablöst"""



"""
Contains Regionskriterium class
"""

from typing import TYPE_CHECKING, Annotated, Literal, Optional

from pydantic import Field

from ..enum.comtyp import ComTyp
from ..utils import postprocess_docstring
from .com import COM

if TYPE_CHECKING:
    from ..enum.gueltigkeitstyp import Gueltigkeitstyp
    from ..enum.regionskriteriumtyp import Regionskriteriumtyp


# pylint: disable=too-few-public-methods


@postprocess_docstring
class Regionskriterium(COM):
    """
    Komponente zur Abbildung eines Regionskriteriums

    .. raw:: html

        <object data="../_static/images/bo4e/com/Regionskriterium.svg" type="image/svg+xml"></object>

    .. HINT::
        `Regionskriterium JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/BO4E/BO4E-Schemas/{__gh_version__}/src/bo4e_schemas/com/Regionskriterium.json>`_

    """

    typ: Annotated[Literal[ComTyp.REGIONSKRITERIUM], Field(alias="_typ")] = ComTyp.REGIONSKRITERIUM

    gueltigkeitstyp: Optional["Gueltigkeitstyp"] = None
    """Hier wird festgelegt, ob es sich um ein einschließendes oder ausschließendes Kriterium handelt."""
    regionskriteriumtyp: Optional["Regionskriteriumtyp"] = None
    """Hier wird das Kriterium selbst angegeben, z.B. Bundesland."""
    wert: Optional[str] = None
    """
    Der Wert, den das Kriterium annehmen kann, z.B. NRW.
    Im Falle des Regionskriteriumstyp BUNDESWEIT spielt dieser Wert keine Rolle.
    """



# pylint: disable=missing-module-docstring
from bo4e.enum.strenum import StrEnum


class Rechnungslegung(StrEnum):
    """
    Aufzählung der Möglichkeiten zur Rechnungslegung in Ausschreibungen.
    """

    MONATSRECHN = "MONATSRECHN"
    """monatsscharfe Rechnung"""
    ABSCHL_MONATSRECHN = "ABSCHL_MONATSRECHN"
    """Abschlag mit Monatsrechnung"""
    ABSCHL_JAHRESRECHN = "ABSCHL_JAHRESRECHN"
    """Abschlag mit Jahresrechnung"""
    MONATSRECHN_JAHRESRECHN = "MONATSRECHN_JAHRESRECHN"
    """Monatsrechnung mit Jahresrechnung"""
    VORKASSE = "VORKASSE"
    """Vorkasse"""



"""
Contains Adresse class
"""

from typing import Annotated, Literal, Optional

from pydantic import Field

from ..enum.comtyp import ComTyp
from ..enum.landescode import Landescode
from ..utils import postprocess_docstring
from .com import COM

# pylint: disable=too-many-instance-attributes, too-few-public-methods


@postprocess_docstring
class Adresse(COM):
    """
    Contains an address that can be used for most purposes.

    .. raw:: html

        <object data="../_static/images/bo4e/com/Adresse.svg" type="image/svg+xml"></object>

    .. HINT::
        `Adresse JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/BO4E/BO4E-Schemas/{__gh_version__}/src/bo4e_schemas/com/Adresse.json>`_

    """

    typ: Annotated[Literal[ComTyp.ADRESSE], Field(alias="_typ")] = ComTyp.ADRESSE

    postleitzahl: Optional[str] = None
    """
    Die Postleitzahl; z.B: "41836"
    """
    ort: Optional[str] = None
    """
    Bezeichnung der Stadt; z.B. "Hückelhoven"
    """

    ortsteil: Optional[str] = None
    """
    Bezeichnung des Ortsteils; z.B. "Mitte"
    """
    strasse: Optional[str] = None
    """
    Bezeichnung der Straße; z.B. "Weserstraße"
    """
    hausnummer: Optional[str] = None
    """
    Hausnummer inkl. Zusatz; z.B. "3", "4a"
    """
    postfach: Optional[str] = None
    """Im Falle einer Postfachadresse das Postfach; Damit werden Straße und Hausnummer nicht berücksichtigt"""
    adresszusatz: Optional[str] = None
    """
    Zusatzhinweis zum Auffinden der Adresse, z.B. "3. Stock linke Wohnung"
    """
    co_ergaenzung: Optional[str] = None
    """
    Im Falle einer c/o-Adresse steht in diesem Attribut die Anrede. Z.B. "c/o Veronica Hauptmieterin"
    """
    landescode: Optional["Landescode"] = Landescode.DE  # type:ignore
    """Offizieller ISO-Landescode"""



from decimal import Decimal

import pytest

from bo4e import AufAbschlagstyp, PositionsAufAbschlag, Waehrungseinheit
from tests.serialization_helper import assert_serialization_roundtrip


class TestPositionsAufAbschlag:
    @pytest.mark.parametrize(
        "positionsaufabschlag",
        [
            pytest.param(
                PositionsAufAbschlag(
                    bezeichnung="foo",
                    beschreibung="bar",
                    auf_abschlagstyp=AufAbschlagstyp.ABSOLUT,
                    auf_abschlagswert=Decimal(4.25),
                    auf_abschlagswaehrung=Waehrungseinheit.EUR,
                ),
            ),
        ],
    )
    def test_serialization_roundtrip(self, positionsaufabschlag: PositionsAufAbschlag) -> None:
        """
        Test de-/serialisation of PositionsAufAbschlag.
        """
        assert_serialization_roundtrip(positionsaufabschlag)



"""
Contains class Aggregationsverantwortungs
"""

from bo4e.enum.strenum import StrEnum


class Aggregationsverantwortung(StrEnum):
    """
    Mögliche Qualifier für die Aggregationsverantwortung
    """

    UENB = "UENB"  #: Übertragungsnetzbetreiber
    VNB = "VNB"  #: Verteilnetzbetreiber



# pylint: disable=missing-module-docstring
from bo4e.enum.strenum import StrEnum


class EMobilitaetsart(StrEnum):
    """
    Art der E-Mobilität
    """

    WALLBOX = "WALLBOX"
    E_MOBILITAETSLADESAEULE = "E_MOBILITAETSLADESAEULE"
    LADEPARK = "LADEPARK"



# pylint:disable=missing-module-docstring

from bo4e.enum.strenum import StrEnum


class Geschaeftspartnerrolle(StrEnum):
    """
    Diese Rollen kann ein Geschäftspartner einnehmen.
    """

    LIEFERANT = "LIEFERANT"
    """Lieferant"""
    DIENSTLEISTER = "DIENSTLEISTER"
    """Dienstleister"""
    KUNDE = "KUNDE"
    """Kunde"""
    INTERESSENT = "INTERESSENT"
    """Interessent"""
    MARKTPARTNER = "MARKTPARTNER"
    """Marktpartner"""



"""
Contains base class for all components
"""

from decimal import Decimal
from typing import Annotated, Optional, Type, TypeVar

from humps.main import camelize

# pylint: disable=no-name-in-module
from pydantic import BaseModel, ConfigDict, Field

from bo4e.version import __version__
from bo4e.zusatzattribut import ZusatzAttribut

# pylint: disable=too-few-public-methods
#
from ..utils import postprocess_docstring


@postprocess_docstring
class COM(BaseModel):
    """
    base class for all components

    .. raw:: html

        <object data="../_static/images/bo4e/com/COM.svg" type="image/svg+xml"></object>

    .. HINT::
        `COM JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/BO4E/BO4E-Schemas/{__gh_version__}/src/bo4e_schemas/com/COM.json>`_

    """

    version: Annotated[Optional[str], Field(alias="_version")] = __version__
    """
    Version der COM-Struktur aka "fachliche Versionierung"
    """

    # Python internal: The field is not named '_id' because leading underscores are not allowed in pydantic field names.
    # NameError: Fields must not use names with leading underscores; e.g., use 'id' instead of '_id'.
    id: Annotated[Optional[str], Field(alias="_id")] = None
    """
    Eine generische ID, die für eigene Zwecke genutzt werden kann.
    Z.B. könnten hier UUIDs aus einer Datenbank stehen oder URLs zu einem Backend-System.
    """

    zusatz_attribute: Optional[list["ZusatzAttribut"]] = None

    # pylint: disable=duplicate-code
    # basic configuration for pydantic's behaviour
    model_config = ConfigDict(
        alias_generator=camelize,
        populate_by_name=True,
        extra="allow",
        # json_encoders is deprecated, but there is no easy-to-use alternative. The best way would be to create
        # an annotated version of Decimal, but you would have to use it everywhere in the pydantic models.
        # See this issue for more info: https://github.com/pydantic/pydantic/issues/6375
        json_encoders={Decimal: str},
        use_attribute_docstrings=True,
    )


# pylint: disable=invalid-name
# Any type derived from COM including those that do not directly inherit from COM
TCom = TypeVar("TCom", bound=Type[COM])


# todo: find out if this way of typing is correct



# pylint: disable=missing-module-docstring
from bo4e.enum.strenum import StrEnum


class Angebotsstatus(StrEnum):
    """
    Gibt den Status eines Angebotes an.
    """

    KONZEPTION = "KONZEPTION"
    """Konzeption"""
    UNVERBINDLICH = "UNVERBINDLICH"
    """unverbindlich"""
    VERBINDLICH = "VERBINDLICH"
    """verbindlich"""
    BEAUFTRAGT = "BEAUFTRAGT"
    """beauftragt"""
    UNGUELTIG = "UNGUELTIG"
    """ungültig"""
    ABGELEHNT = "ABGELEHNT"
    """abgelehnt"""
    NACHGEFASST = "NACHGEFASST"
    """nachgefasst"""
    AUSSTEHEND = "AUSSTEHEND"
    """ausstehend"""
    ERLEDIGT = "ERLEDIGT"
    """erledigt"""



import pytest

from bo4e import Adresse, Ausschreibungsdetail, Menge, Netzebene, Zaehlertyp, Zeitraum
from tests.serialization_helper import assert_serialization_roundtrip


class TestAusschreibungsdetail:
    @pytest.mark.parametrize(
        "ausschreibungsdetail",
        [
            pytest.param(
                Ausschreibungsdetail(
                    marktlokations_id="56789012345",
                    netzebene_lieferung=Netzebene.MSP,
                    netzebene_messung=Netzebene.NSP,
                    marktlokationsadresse=Adresse(),
                    lieferzeitraum=Zeitraum(),
                    netzbetreiber="Stromnetz Hamburg GmbH",
                    kunde="Dei Mudder ihr Kunde",
                    zaehlernummer="1YSK4234092304",
                    marktlokationsbezeichnung="Zentraler Einkauf, Hamburg",
                    zaehlertechnik=Zaehlertyp.LEISTUNGSZAEHLER,
                    ist_lastgang_vorhanden=True,
                    prognose_jahresarbeit=Menge(),
                    prognose_arbeit_lieferzeitraum=Menge(),
                    prognose_leistung=Menge(),
                    rechnungsadresse=Adresse(),
                ),
                id="all attributes at first level",
            ),
        ],
    )
    def test_serialization_roundtrip(self, ausschreibungsdetail: Ausschreibungsdetail) -> None:
        """
        Test de-/serialisation of Ausschreibungsdetail
        """
        assert_serialization_roundtrip(ausschreibungsdetail)



# pylint: disable=missing-module-docstring
from bo4e.enum.strenum import StrEnum


class Kundentyp(StrEnum):
    """
    Auflistung der Typen von Endkunden. Daraus kann das Verbrauchsprofil abgeleitet werden.
    """

    GEWERBE = "GEWERBE"
    """Gewerbekunden"""
    PRIVAT = "PRIVAT"
    """Privatkunden"""
    LANDWIRT = "LANDWIRT"
    """Landwirte"""
    SONSTIGE = "SONSTIGE"
    """Sonstige Endkunden"""
    HAUSHALT = "HAUSHALT"
    """Haushaltskunden"""
    DIREKTHEIZUNG = "DIREKTHEIZUNG"
    """Direktheizungen"""
    GEMEINSCHAFT_MFH = "GEMEINSCHAFT_MFH"
    """Gemeinschaftseinrichtungen von MFH"""
    KIRCHE = "KIRCHE"
    """Kirchen und caritative Einrichtungen"""
    KWK = "KWK"
    """KWK-Anlagen"""
    LADESAEULE = "LADESAEULE"
    """Ladesäulen"""
    BELEUCHTUNG_OEFFENTLICH = "BELEUCHTUNG_OEFFENTLICH"
    """Öffentliche Beleuchtungen"""
    BELEUCHTUNG_STRASSE = "BELEUCHTUNG_STRASSE"
    """Straßenbeleuchtungen"""
    SPEICHERHEIZUNG = "SPEICHERHEIZUNG"
    """Speicherheizungen"""
    UNTERBR_EINRICHTUNG = "UNTERBR_EINRICHTUNG"
    """Unterbrechbare Verbrauchseinrichtungen"""
    WAERMEPUMPE = "WAERMEPUMPE"
    """Wärmepumpen"""



"""
Der ISO-Landescode als Enumeration.
"""

from enum import Enum

from iso3166 import countries

alpha2codes = {c.alpha2: c.alpha2 for c in countries}

Landescode = Enum("Landescode", alpha2codes)  # type: ignore



"""
Contains Dienstleistung class
"""

from typing import TYPE_CHECKING, Annotated, Literal, Optional

from pydantic import Field

from ..enum.comtyp import ComTyp
from ..utils import postprocess_docstring
from .com import COM

if TYPE_CHECKING:
    from ..enum.dienstleistungstyp import Dienstleistungstyp


# pylint: disable=too-few-public-methods


@postprocess_docstring
class Dienstleistung(COM):
    """
    Abbildung einer abrechenbaren Dienstleistung.

    .. raw:: html

        <object data="../_static/images/bo4e/com/Dienstleistung.svg" type="image/svg+xml"></object>

    .. HINT::
        `Dienstleistung JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/BO4E/BO4E-Schemas/{__gh_version__}/src/bo4e_schemas/com/Dienstleistung.json>`_

    """

    typ: Annotated[Literal[ComTyp.DIENSTLEISTUNG], Field(alias="_typ")] = ComTyp.DIENSTLEISTUNG

    dienstleistungstyp: Optional["Dienstleistungstyp"] = None
    """Kennzeichnung der Dienstleistung"""
    bezeichnung: Optional[str] = None
    """Bezeichnung der Dienstleistung"""



from __future__ import annotations

import pytest

from bo4e import Anrede, Organisationstyp, Titel
from bo4e.bo.geschaeftspartner import Geschaeftspartner
from bo4e.bo.person import Person
from bo4e.com.adresse import Adresse
from bo4e.enum.geschaeftspartnerrolle import Geschaeftspartnerrolle
from bo4e.zusatzattribut import ZusatzAttribut
from tests.serialization_helper import assert_serialization_roundtrip


class TestZusatzAttribut:
    @pytest.mark.parametrize(
        "geschaeftspartner",
        [
            pytest.param(
                Geschaeftspartner(
                    zusatz_attribute=[
                        ZusatzAttribut(name="SAP GP Nummer", wert="0123456789"),
                        ZusatzAttribut(name="Schufa-ID", wert="aksdlakoeuhn"),
                    ],
                    # just some dummy data to make the GP valid
                    ansprechpartner=[Person(), Person()],
                    anrede=Anrede.EHELEUTE,
                    individuelle_anrede="Künstler",
                    titel=Titel.PROF_DR,
                    vorname="Hans",
                    nachname="Müller-Schmidt",
                    organisationstyp=Organisationstyp.PRIVATPERSON,
                    geschaeftspartnerrollen=[Geschaeftspartnerrolle.KUNDE],
                    adresse=Adresse(
                        strasse="Am Geldspeicher",
                        hausnummer="17",
                        postleitzahl="88101",
                        ort="Entenhausen",
                    ),
                ),
            ),
        ],
    )
    def test_serialization_roundtrip(self, geschaeftspartner: Geschaeftspartner) -> None:
        """
        Test de-/serialisation of Geschaeftspartner.
        """
        assert_serialization_roundtrip(geschaeftspartner)



"""
Contains Zaehlzeitregister class
"""

from typing import Annotated, Literal, Optional

from pydantic import Field

from ..enum.comtyp import ComTyp
from ..utils import postprocess_docstring
from .com import COM

# pylint: disable=no-name-in-module
# pylint: disable=too-few-public-methods


@postprocess_docstring
class Zaehlzeitregister(COM):
    """
    Mit dieser Komponente werden Zählzeitregister modelliert. Ein Zählzeitregister beschreibt eine erweiterte Definition der Zählzeit
    in Bezug auf ein Register. Dabei werden alle Codes dazu vom Netzbetreiber vergeben.

    .. raw:: html

        <object data="../_static/images/bo4e/com/Zaehlzeitregister.svg" type="image/svg+xml"></object>

    .. HINT::
        `Zaehlzeitregister JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/BO4E/BO4E-Schemas/{__gh_version__}/src/bo4e_schemas/com/Zaehlzeitregister.json>`_

    """

    typ: Annotated[Literal[ComTyp.ZAEHLZEITREGISTER], Field(alias="_typ")] = ComTyp.ZAEHLZEITREGISTER

    zaehlzeit_definition: Optional[str] = None
    """Zählzeitdefinition"""
    zaehlzeit_register: Optional[str] = None
    """Zählzeitregister"""
    ist_schwachlastfaehig: Optional[bool] = None
    """Schwachlastfaehigkeit"""



"""
Contains Verbrauch
"""

from decimal import Decimal
from typing import TYPE_CHECKING, Annotated, Literal, Optional

import pydantic
from pydantic import Field

from ..enum.comtyp import ComTyp
from ..utils import postprocess_docstring
from .com import COM

if TYPE_CHECKING:
    from ..enum.mengeneinheit import Mengeneinheit
    from ..enum.messwertstatus import Messwertstatus


# pylint: disable=too-few-public-methods
# pylint: disable=no-name-in-module


@postprocess_docstring
class Verbrauch(COM):
    """
    Abbildung eines zeitlich abgegrenzten Verbrauchs

    .. raw:: html

        <object data="../_static/images/bo4e/com/Verbrauch.svg" type="image/svg+xml"></object>

    .. HINT::
        `Verbrauch JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/BO4E/BO4E-Schemas/{__gh_version__}/src/bo4e_schemas/com/Verbrauch.json>`_

    """

    typ: Annotated[Literal[ComTyp.VERBRAUCH], Field(alias="_typ")] = ComTyp.VERBRAUCH

    obis_kennzahl: Optional[str] = None
    """Die OBIS-Kennzahl für den Wert, die festlegt, welche Größe mit dem Stand gemeldet wird, z.B. '1-0:"""
    wert: Optional[Decimal] = None
    """Gibt den absoluten Wert der Menge an"""
    einheit: Optional["Mengeneinheit"] = None
    """Gibt die Einheit zum jeweiligen Wert an"""

    startdatum: Optional[pydantic.AwareDatetime] = None
    """Inklusiver Beginn des Zeitraumes, für den der Verbrauch angegeben wird"""
    enddatum: Optional[pydantic.AwareDatetime] = None
    """
    Exklusives Ende des Zeitraumes, für den der Verbrauch angegeben wird.
    Ein ZeitPUNKT kann mit start==ende angegeben werden.
    """
    messwertstatus: Optional["Messwertstatus"] = None
    """Messwertstatus includes the plausibility of the value"""



from decimal import Decimal

import pytest

from bo4e import AufAbschlagstaffelProOrt
from tests.serialization_helper import assert_serialization_roundtrip


class TestAufAbschlagstaffelProOrt:
    @pytest.mark.parametrize(
        "aufabschlagstaffelproort",
        [
            pytest.param(
                AufAbschlagstaffelProOrt(
                    wert=Decimal(2.5),
                    staffelgrenze_von=Decimal(1),
                    staffelgrenze_bis=Decimal(5),
                ),
            ),
        ],
    )
    def test_aufabschlagstaffelproort_required_attributes(
        self, aufabschlagstaffelproort: AufAbschlagstaffelProOrt
    ) -> None:
        """
        Test de-/serialisation of AufAbschlagstaffelProOrt.
        """
        assert_serialization_roundtrip(aufabschlagstaffelproort)



import pytest

from bo4e import Gueltigkeitstyp, KriteriumWert, RegionaleGueltigkeit, Tarifregionskriterium
from tests.serialization_helper import assert_serialization_roundtrip


class TestRegionaleGueltigkeit:
    @pytest.mark.parametrize(
        "regionalegueltigkeit",
        [
            pytest.param(
                RegionaleGueltigkeit(
                    gueltigkeitstyp=Gueltigkeitstyp.NUR_IN,
                    kriteriums_werte=[
                        KriteriumWert(
                            kriterium=Tarifregionskriterium.NETZ_NUMMER,
                            wert="12345",
                        ),
                    ],
                ),
                id="all attributes at first level",
            ),
        ],
    )
    def test_regionalegueltigkeit_serialization_roundtrip(self, regionalegueltigkeit: RegionaleGueltigkeit) -> None:
        """
        Test de-/serialisation of RegionaleGueltigkeit with minimal attributes.
        """
        assert_serialization_roundtrip(regionalegueltigkeit)



# pylint: disable=missing-module-docstring
from bo4e.enum.strenum import StrEnum


class Bilanzierungsmethode(StrEnum):
    """
    Mit dieser Aufzählung kann zwischen den Bilanzierungsmethoden bzw. -grundlagen unterschieden werden.
    """

    RLM = "RLM"
    """Registrierende Leistungsmessung"""
    SLP = "SLP"
    """Standard Lastprofil"""
    TLP_GEMEINSAM = "TLP_GEMEINSAM"
    """TLP gemeinsame Messung"""
    TLP_GETRENNT = "TLP_GETRENNT"
    """TLP getrennte Messung"""
    PAUSCHAL = "PAUSCHAL"
    """Pauschale Betrachtung (Band)"""
    IMS = "IMS"
    """Intelligentes Messsystem"""



# pylint: disable=missing-module-docstring
from bo4e.enum.strenum import StrEnum


# pylint:disable=empty-docstring
# no docstring in official docs as of 2021-12-01
class Leistungstyp(StrEnum):
    """ """

    ARBEITSPREIS_WIRKARBEIT = "ARBEITSPREIS_WIRKARBEIT"
    """Arbeitspreis zur Abrechnung der Wirkarbeit"""
    LEISTUNGSPREIS_WIRKLEISTUNG = "LEISTUNGSPREIS_WIRKLEISTUNG"
    """Leistungspreis zur Abrechnung der Wirkleistung"""
    ARBEITSPREIS_BLINDARBEIT_IND = "ARBEITSPREIS_BLINDARBEIT_IND"
    """Arbeitspreis zur Abrechnung der Blindarbeit induktiv"""
    ARBEITSPREIS_BLINDARBEIT_KAP = "ARBEITSPREIS_BLINDARBEIT_KAP"
    """Arbeitspreis zur Abrechnung der Blindarbeit kapazitiv"""
    GRUNDPREIS = "GRUNDPREIS"
    """Grundpreis (pro Zeiteinheit)"""
    GRUNDPREIS_ARBEIT = "GRUNDPREIS_ARBEIT"
    """Grundpreis, der auf die Arbeit berechnet wird (bei RLM)"""
    GRUNDPREIS_LEISTUNG = "GRUNDPREIS_LEISTUNG"
    """Grundpreis, der auf die Leistung berechnet wird (bei RLM)"""
    MEHRMINDERMENGE = "MEHRMINDERMENGE"
    """Mehr- oder Mindermenge"""
    MESSSTELLENBETRIEB = "MESSSTELLENBETRIEB"
    """Preis pro Zeiteinheit"""
    MESSDIENSTLEISTUNG = "MESSDIENSTLEISTUNG"
    """Preis pro Zeiteinheit"""
    MESSDIENSTLEISTUNG_INKL_MESSUNG = "MESSDIENSTLEISTUNG_INKL_MESSUNG"
    """MDL inklusive der Messung (ab 2017), Preis pro Zeiteinheit"""
    ABRECHNUNG = "ABRECHNUNG"
    """Preis pro Zeiteinheit"""
    KONZESSIONS_ABGABE = "KONZESSIONS_ABGABE"
    """Konzessionsabgabe"""
    KWK_UMLAGE = "KWK_UMLAGE"
    """KWK-Umlage"""
    OFFSHORE_UMLAGE = "OFFSHORE_UMLAGE"
    """Offshore-Haftungsumlage"""
    ABLAV_UMLAGE = "ABLAV_UMLAGE"
    """Umlage für abschaltbare Lasten"""
    SONDERKUNDEN_UMLAGE = "SONDERKUNDEN_UMLAGE"
    """§19 StromNEV Umlage"""
    REGELENERGIE_UMLAGE = "REGELENERGIE_UMLAGE"
    """Regelenergieumlage"""
    BILANZIERUNG_UMLAGE = "BILANZIERUNG_UMLAGE"
    """Bilanzierungsumlage"""
    AUSLESUNG_ZUSAETZLICH = "AUSLESUNG_ZUSAETZLICH"
    """Zusätzliche Auslesung (pro Vorgang)"""
    ABLESUNG_ZUSAETZLICH = "ABLESUNG_ZUSAETZLICH"
    """Zusätzliche Ablesung (pro Vorgang)"""
    ABRECHNUNG_ZUSAETZLICH = "ABRECHNUNG_ZUSAETZLICH"
    """Zusätzliche Abresung (pro Vorgang)"""
    SPERRUNG = "SPERRUNG"
    """Sperrung einer Abnahmestelle"""
    ENTSPERRUNG = "ENTSPERRUNG"
    """Entsperrung einer Abnahmestelle"""
    MAHNKOSTEN = "MAHNKOSTEN"
    """Mahnkosten"""
    INKASSOKOSTEN = "INKASSOKOSTEN"
    """Inkassokosten"""
    EEG_UMLAGE = "EEG_UMLAGE"
    """EEG-Umlage"""
    ENERGIESTEUER = "ENERGIESTEUER"
    """Strom- oder Erdgassteuer"""
    NETZPREIS = "NETZPREIS"
    """Netzpreis"""
    MESSPREIS = "MESSPREIS"
    """Messpreis"""
    SONSTIGER_PREIS = "SONSTIGER_PREIS"
    """Sonstiger_Preis"""
    DIENSTLEISTUNG = "DIENSTLEISTUNG"
    """Dienstleistung"""



from decimal import Decimal

import pytest

from bo4e import TarifpreisstaffelProOrt
from tests.serialization_helper import assert_serialization_roundtrip


class TestTarifpreisstaffelProOrt:
    @pytest.mark.parametrize(
        "tarifpreisstaffelproort",
        [
            pytest.param(
                TarifpreisstaffelProOrt(
                    arbeitspreis=Decimal(10),
                    arbeitspreis_n_t=Decimal(11),
                    grundpreis=Decimal(12),
                    staffelgrenze_von=Decimal(13),
                    staffelgrenze_bis=Decimal(14),
                ),
                id="maximal (and minimal) attributes",
            ),
        ],
    )
    def test_serialization_roundtrip(self, tarifpreisstaffelproort: TarifpreisstaffelProOrt) -> None:
        """
        Test de-/serialisation
        """
        assert_serialization_roundtrip(tarifpreisstaffelproort)



# pylint: disable=missing-module-docstring
from bo4e.enum.strenum import StrEnum


class Preismodell(StrEnum):
    """
    Bezeichnung der Preismodelle in Ausschreibungen für die Energielieferung.
    """

    FESTPREIS = "FESTPREIS"
    """FESTPREIS"""
    TRANCHE = "TRANCHE"
    """TRANCHE"""



"""
This script is run in the tox 'json_schemas' environment.
"""

import importlib
import inspect
import json
import logging
import pkgutil
import re
import sys
from enum import Enum
from pathlib import Path
from typing import Any, Iterator, Literal, cast

import click
from bo4e_cli.io.version_file import create_version_file
from bo4e_cli.models.version import Version
from pydantic import BaseModel, TypeAdapter
from pydantic.json_schema import GenerateJsonSchema as _GenerateJsonSchema
from pydantic.json_schema import JsonSchemaValue
from pydantic_core import core_schema

from bo4e import ZusatzAttribut, __gh_version__

logging.basicConfig(level=logging.INFO, stream=sys.stdout)
_logger = logging.getLogger(__name__)
root_directory = Path(__file__).parent
output_directory = root_directory / "json_schemas"

NEW_REF_TEMPLATE = "https://raw.githubusercontent.com/BO4E/BO4E-Schemas/{version}/src/bo4e_schemas/{pkg}/{model}.json"
NEW_REF_TEMPLATE_ROOT = "https://raw.githubusercontent.com/BO4E/BO4E-Schemas/{version}/src/bo4e_schemas/{model}.json"
OLD_REF_TEMPLATE = re.compile(r"^#/\$defs/(?P<model>\w+)$")

PARSABLE_CLASS_TYPE = type[BaseModel] | type[Enum]


class GenerateJsonSchema(_GenerateJsonSchema):
    """
    This class is a copy of pydantic.json_schema.GenerateJsonSchema with the only difference that the
    decimal_schema method is overwritten to generate a JSON schema that can be used by the
    BO4E-Python-Generator (https://github.com/bo4e/BO4E-Python-Generator).
    """

    def decimal_schema(self, schema: core_schema.DecimalSchema) -> JsonSchemaValue:
        """
        Generates a JSON schema that matches a decimal value.
        The output format is changed to work well with BO4E-Python-Generator.
        """
        json_schema = self.float_schema(core_schema.float_schema())
        if self.mode == "validation":
            json_schema["format"] = "decimal"
        return json_schema


def delete_json_schemas(packages: list[str]) -> None:
    """delete all json schemas"""
    for pkg in packages:
        to_delete = output_directory / pkg
        if to_delete.exists():
            for file in to_delete.iterdir():
                file.unlink()


def get_models(pkg: str) -> Iterator[str]:
    """
    Get all models in a package
    """
    yield from (name for _, name, _ in pkgutil.iter_modules([str(root_directory / "src" / "bo4e" / pkg)]))


def get_classes(modl_name: str) -> list[tuple[str, type]]:
    """
    Get all classes in a module
    """
    modl = importlib.import_module(modl_name)
    return inspect.getmembers(modl, lambda member: inspect.isclass(member) and member.__module__ == modl_name)


def get_namespace(packages: list[str]) -> dict[str, tuple[str, str, PARSABLE_CLASS_TYPE]]:
    """
    Builds a dictionary with the classnames as keys and their module as tuples in the values. E.g.:
    {
        "Angebot": ("bo", "angebot"),
        "Adresse": ("com", "adresse"),
        ...
    }
    This function filters out all classes which names begin with an underscore.
    """
    namespace = {}
    for pkg in packages:
        for model in get_models(pkg):
            modl_name = f"bo4e.{pkg}.{model}"
            cls_list = get_classes(modl_name)
            for name, cls in cls_list:
                if not name.startswith("_") and name != "StrEnum":
                    namespace[name] = (pkg, model, cast(PARSABLE_CLASS_TYPE, cls))
    return namespace


def get_schema_json_dict(cls: Any) -> dict[str, Any]:
    """
    Get the json schema for a class
    """
    if issubclass(cls, BaseModel):
        schema_json_dict = cls.model_json_schema(schema_generator=GenerateJsonSchema)
    elif issubclass(cls, Enum):
        schema_json_dict = TypeAdapter(cls).json_schema(schema_generator=GenerateJsonSchema)
    else:
        raise ValueError(f"Class {cls} is neither a pydantic BaseModel nor an enum.")
    if {"allOf", "$defs"} == set(schema_json_dict.keys()):
        assert (
            len(schema_json_dict["allOf"]) == 1
        ), "Internal error: Assumed circular reference but structure is unexpected"
        # This is the case for schemas containing circular references
        reference_pattern = re.compile(r"^#/\$defs/(?P<cls_name>\w+)$")
        reference_match = reference_pattern.fullmatch(schema_json_dict["allOf"][0]["$ref"])
        assert (
            reference_match is not None
        ), f"Internal Error: Reference string has unexpected format: {schema_json_dict['allOf'][0]['$ref']}"
        schema_json_dict_to_merge = schema_json_dict["$defs"][reference_match.group("cls_name")]
        del schema_json_dict["allOf"]
        schema_json_dict.update(schema_json_dict_to_merge)
    if {"$ref", "$defs"} == set(schema_json_dict.keys()):
        # The newer version of pydantic sometimes generates a schema with only a $ref and a $defs key where the $ref
        # field points to the actual schema definition in the $defs field.
        reference_pattern = re.compile(r"^#/\$defs/(?P<cls_name>\w+)$")
        reference_match = reference_pattern.fullmatch(schema_json_dict["$ref"])
        assert (
            reference_match is not None
        ), f"Internal Error: Reference string has unexpected format: {schema_json_dict['$ref']}"
        schema_json_dict_to_merge = schema_json_dict["$defs"][reference_match.group("cls_name")]
        del schema_json_dict["$ref"]
        schema_json_dict.update(schema_json_dict_to_merge)
    if "$defs" in schema_json_dict:
        del schema_json_dict["$defs"]
    return schema_json_dict


def validate_schema(file_path: Path, schema_json_dict: dict[str, Any], name: str) -> None:
    """
    Validate the schema for a class
    """
    with open(file_path, "r", encoding="utf-8") as json_schema_file:
        existing_schema = json.load(json_schema_file)

    if schema_json_dict != existing_schema:
        raise ValueError(f"Schema for {name} has changed. Please run this script with mode 'generate'.")
        # or call tox -e generate_json_schemas
    _logger.debug("Schema for %s is consistent", name)


def generate_schema(file_path: Path, schema_json_dict: dict[str, Any]) -> None:
    """
    Generate the schema for a class
    """
    if not file_path.parent.exists():
        file_path.parent.mkdir(parents=True)
    with open(file_path, "w+", encoding="utf-8") as json_schema_file:
        json_schema_file.write(json.dumps(schema_json_dict, indent=4, sort_keys=True, ensure_ascii=False))
        json_schema_file.write("\n")


def generate_version_file(file_path: Path, version: str) -> None:
    """
    Generate a version file with the given version
    """
    with open(file_path, "w+", encoding="utf-8") as version_file:
        version_file.write(version)
    _logger.info("Generated version file at %s with content %s", file_path, version)


def replace_refs(
    schema_json_dict: dict[str, Any], namespace: dict[str, tuple[str, str, PARSABLE_CLASS_TYPE]], target_version: str
) -> None:
    """
    Replace the definition of a class with an online reference to the definition
    """

    def traverse_list(obj: list[Any]) -> None:
        for item in obj:
            if isinstance(item, dict):
                traverse_dict(item)
            elif isinstance(item, list):
                traverse_list(item)

    def traverse_dict(obj: dict[str, Any]) -> None:
        for key, value in obj.items():
            if isinstance(value, dict):
                traverse_dict(value)
            elif isinstance(value, list):
                traverse_list(value)
            elif key == "$ref":
                match = OLD_REF_TEMPLATE.match(value)
                if match is None:
                    raise ValueError(f"Invalid reference: {value}")
                ref_model = match.group("model")
                if ref_model not in namespace:
                    raise ValueError(f"Unknown referenced model {ref_model}")
                if namespace[ref_model][0] == "":
                    obj["$ref"] = NEW_REF_TEMPLATE_ROOT.format(model=ref_model, version=target_version)
                else:
                    obj["$ref"] = NEW_REF_TEMPLATE.format(
                        pkg=namespace[ref_model][0], model=ref_model, version=target_version
                    )

    traverse_dict(schema_json_dict)


@click.command()
@click.option(
    "--mode",
    "-m",
    help="use 'validate' to validate existing schemas or 'generate' to generate new schemas",
    required=True,
    type=click.Choice(["validate", "generate"]),
)
@click.option(
    "--target-version",
    "-t",
    help="the tagged version known inside the release workflow",
    required=False,
    type=click.STRING,
    envvar="TARGET_VERSION",
    default=None,
)
def generate_or_validate_json_schemas(mode: Literal["validate", "generate"], target_version: str | None) -> None:
    """generate json schemas for all BOs and COMs"""
    version = Version.from_str(target_version or __gh_version__)
    _logger.info("Mode: %s, target version: %s", mode, version)
    packages = ["bo", "com", "enum"]

    if mode == "generate":
        delete_json_schemas(packages)

    namespace = get_namespace(packages)
    namespace[ZusatzAttribut.__name__] = ("", "zusatzattribut", ZusatzAttribut)
    del namespace["Geschaeftsobjekt"]
    del namespace["COM"]

    for name, (pkg, _, cls) in namespace.items():
        _logger.debug("Processing %s", name)

        if pkg == "":
            file_path = output_directory / (name + ".json")
        else:
            file_path = output_directory / pkg / (name + ".json")

        schema_json_dict = get_schema_json_dict(cls)
        replace_refs(schema_json_dict, namespace, str(version))

        if mode == "validate":
            validate_schema(file_path, schema_json_dict, name)
        elif mode == "generate":
            generate_schema(file_path, schema_json_dict)
            _logger.info("Generated schema for %s", name)
        else:
            raise ValueError(f"Unknown mode '{mode}'")
    if mode == "generate":
        create_version_file(output_directory, version)
        _logger.info("Generated version file at %s with version %s", output_directory / ".version", version)


if __name__ == "__main__":
    generate_or_validate_json_schemas()  # pylint:disable=no-value-for-parameter



"""
Contains Preisblatt class
"""

# pylint: disable=too-few-public-methods
# pylint: disable=no-name-in-module
from typing import TYPE_CHECKING, Annotated, Literal, Optional

from pydantic import Field

from ..enum.botyp import BoTyp
from ..utils import postprocess_docstring
from .geschaeftsobjekt import Geschaeftsobjekt

if TYPE_CHECKING:
    from ..com.preisposition import Preisposition
    from ..com.zeitraum import Zeitraum
    from ..enum.preisstatus import Preisstatus
    from ..enum.sparte import Sparte
    from .marktteilnehmer import Marktteilnehmer


@postprocess_docstring
class Preisblatt(Geschaeftsobjekt):
    """
    Das allgemeine Modell zur Abbildung von Preisen;
    Davon abgeleitet können, über die Zuordnung identifizierender Merkmale, spezielle Preisblatt-Varianten modelliert
    werden.

    Die jeweiligen Sätze von Merkmalen sind in der Grafik ergänzt worden und stellen jeweils eine Ausprägung für die
    verschiedenen Anwendungsfälle der Preisblätter dar.

    .. raw:: html

        <object data="../_static/images/bo4e/bo/Preisblatt.svg" type="image/svg+xml"></object>

    .. HINT::
        `Preisblatt JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/BO4E/BO4E-Schemas/{__gh_version__}/src/bo4e_schemas/bo/Preisblatt.json>`_

    """

    typ: Annotated[Literal[BoTyp.PREISBLATT], Field(alias="_typ")] = BoTyp.PREISBLATT
    bezeichnung: Optional[str] = None
    """Eine Bezeichnung für das Preisblatt"""
    sparte: Optional["Sparte"] = None
    """Preisblatt gilt für angegebene Sparte"""
    preisstatus: Optional["Preisstatus"] = None
    """Merkmal, das anzeigt, ob es sich um vorläufige oder endgültige Preise handelt"""
    gueltigkeit: Optional["Zeitraum"] = None
    """Der Zeitraum für den der Preis festgelegt ist"""
    preispositionen: Optional[list["Preisposition"]] = None
    """Die einzelnen Positionen, die mit dem Preisblatt abgerechnet werden können. Z.B. Arbeitspreis, Grundpreis etc"""
    herausgeber: Optional["Marktteilnehmer"] = None
    """Der Netzbetreiber, der die Preise veröffentlicht hat"""



# pylint: disable=missing-module-docstring
from bo4e.enum.strenum import StrEnum


class Tarifmerkmal(StrEnum):
    """
    Produktmerkmale im Zusammenhang mit der Tarifdefinition.
    """

    STANDARD = "STANDARD"
    """Standardprodukt"""
    VORKASSE = "VORKASSE"
    """Vorkassenprodukt"""
    PAKET = "PAKET"
    """Paketpreisprodukt"""
    KOMBI = "KOMBI"
    """Kombiprodukt"""
    FESTPREIS = "FESTPREIS"
    """Festpreisprodukt"""
    BAUSTROM = "BAUSTROM"
    """Baustromprodukt"""
    HAUSLICHT = "HAUSLICHT"
    """Hauslichtprodukt"""
    HEIZSTROM = "HEIZSTROM"
    """Heizstromprodukt"""
    ONLINE = "ONLINE"
    """Onlineprodukt"""



"""
Contains Rechnung class
"""

from typing import TYPE_CHECKING, Annotated, Literal, Optional

import pydantic
from pydantic import Field

from ..enum.botyp import BoTyp
from ..utils import postprocess_docstring
from .geschaeftsobjekt import Geschaeftsobjekt

if TYPE_CHECKING:
    from ..com.betrag import Betrag
    from ..com.rechnungsposition import Rechnungsposition
    from ..com.steuerbetrag import Steuerbetrag
    from ..com.zeitraum import Zeitraum
    from ..enum.netznutzungrechnungsart import NetznutzungRechnungsart
    from ..enum.netznutzungrechnungstyp import NetznutzungRechnungstyp
    from ..enum.rechnungsstatus import Rechnungsstatus
    from ..enum.rechnungstyp import Rechnungstyp
    from ..enum.sparte import Sparte
    from .geschaeftspartner import Geschaeftspartner
    from .marktlokation import Marktlokation
    from .messlokation import Messlokation

# pylint: disable=too-few-public-methods, too-many-instance-attributes


@postprocess_docstring
class Rechnung(Geschaeftsobjekt):
    """
    Modell für die Abbildung von Rechnungen und Netznutzungsrechnungen im Kontext der Energiewirtschaft;

    .. raw:: html

        <object data="../_static/images/bo4e/bo/Rechnung.svg" type="image/svg+xml"></object>

    .. HINT::
        `Rechnung JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/BO4E/BO4E-Schemas/{__gh_version__}/src/bo4e_schemas/bo/Rechnung.json>`_

    """

    typ: Annotated[Literal[BoTyp.RECHNUNG], Field(alias="_typ")] = BoTyp.RECHNUNG
    ist_storno: Optional[bool] = None
    """
    Kennzeichnung, ob es sich um eine Stornorechnung handelt;
    im Falle "true" findet sich im Attribut "originalrechnungsnummer" die Nummer der Originalrechnung.
    """
    rechnungsnummer: Optional[str] = None
    """Eine im Verwendungskontext eindeutige Nummer für die Rechnung"""
    rechnungsdatum: Optional[pydantic.AwareDatetime] = None
    """Ausstellungsdatum der Rechnung"""
    faelligkeitsdatum: Optional[pydantic.AwareDatetime] = None
    """Zu diesem Datum ist die Zahlung fällig"""
    rechnungstyp: Optional["Rechnungstyp"] = None
    """Ein kontextbezogender Rechnungstyp, z.B. Netznutzungsrechnung"""
    rechnungsperiode: Optional["Zeitraum"] = None
    """Der Zeitraum der zugrunde liegenden Lieferung zur Rechnung"""
    rechnungsersteller: Optional["Geschaeftspartner"] = None
    """Der Aussteller der Rechnung, die Rollencodenummer kennt man über den im Geschäftspartner verlinkten Marktteilnehmer"""
    rechnungsempfaenger: Optional["Geschaeftspartner"] = None
    """Der Aussteller der Rechnung, die Rollencodenummer kennt man über den im Geschäftspartner verlinkten Marktteilnehmer"""
    gesamtnetto: Optional["Betrag"] = None
    """Die Summe der Nettobeträge der Rechnungsteile"""
    gesamtsteuer: Optional["Betrag"] = None
    """Die Summe der Steuerbeträge der Rechnungsteile"""
    gesamtbrutto: Optional["Betrag"] = None
    """Die Summe aus Netto- und Steuerbetrag"""
    zu_zahlen: Optional["Betrag"] = None
    """Der zu zahlende Betrag, der sich aus (gesamtbrutto - vorausbezahlt - rabattBrutto) ergibt"""
    rechnungspositionen: Optional[list["Rechnungsposition"]] = None
    """Die Rechnungspositionen"""
    rechnungstitel: Optional[str] = None
    """Bezeichnung für die vorliegende Rechnung"""
    rechnungsstatus: Optional["Rechnungsstatus"] = None
    """Status der Rechnung zur Kennzeichnung des Bearbeitungsstandes"""
    original_rechnungsnummer: Optional[str] = None
    """Im Falle einer Stornorechnung (storno = true) steht hier die Rechnungsnummer der stornierten Rechnung"""
    vorausgezahlt: Optional["Betrag"] = None
    """Die Summe evtl. vorausgezahlter Beträge, z.B. Abschläge. Angabe als Bruttowert"""
    rabatt_brutto: Optional["Betrag"] = None
    """Gesamtrabatt auf den Bruttobetrag"""
    steuerbetraege: Optional[list["Steuerbetrag"]] = None
    """
    Eine Liste mit Steuerbeträgen pro Steuerkennzeichen/Steuersatz;
    die Summe dieser Beträge ergibt den Wert für gesamtsteuer.
    """
    sparte: Optional["Sparte"] = None
    """Sparte (Strom, Gas ...) für die die Rechnung ausgestellt ist"""
    netznutzungrechnungsart: Optional["NetznutzungRechnungsart"] = None
    """Aus der INVOIC entnommen, befüllt wenn es sich um eine Netznutzungsrechnung handelt"""
    netznutzungrechnungstyp: Optional["NetznutzungRechnungstyp"] = None
    """Aus der INVOIC entnommen, befüllt wenn es sich um eine Netznutzungsrechnung handelt"""
    ist_original: Optional[bool] = None
    """Kennzeichen, ob es sich um ein Original (true) oder eine Kopie handelt (false)"""
    ist_simuliert: Optional[bool] = None
    """Kennzeichen, ob es sich um eine simulierte Rechnung, z.B. zur Rechnungsprüfung handelt"""
    marktlokation: Optional["Marktlokation"] = None
    """Marktlokation, auf die sich die Rechnung bezieht"""
    messlokation: Optional["Messlokation"] = None
    """Messlokation, auf die sich die Rechnung bezieht"""



"""
Contains Geraet class
"""

from typing import TYPE_CHECKING, Annotated, Literal, Optional

from pydantic import Field

from ..enum.botyp import BoTyp
from ..utils import postprocess_docstring
from .geschaeftsobjekt import Geschaeftsobjekt

if TYPE_CHECKING:
    from ..enum.geraeteklasse import Geraeteklasse
    from ..enum.geraetetyp import Geraetetyp


# pylint: disable=too-few-public-methods


@postprocess_docstring
class Geraet(Geschaeftsobjekt):
    """
    Mit diesem BO werden alle Geräte modelliert, die keine Zähler sind.

    .. raw:: html

        <object data="../_static/images/bo4e/bo/Geraet.svg" type="image/svg+xml"></object>

    .. HINT::
        `Geraet JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/BO4E/BO4E-Schemas/{__gh_version__}/src/bo4e_schemas/bo/Geraet.json>`_

    """

    typ: Annotated[Literal[BoTyp.GERAET], Field(alias="_typ")] = BoTyp.GERAET

    geraetenummer: Optional[str] = None
    """Die auf dem Gerät aufgedruckte Nummer, die vom MSB vergeben wird."""
    bezeichnung: Optional[str] = None
    """Bezeichnung des Geräts"""
    geraeteklasse: Optional["Geraeteklasse"] = None
    """Die übergreifende Klasse eines Geräts, beispielsweise Wandler"""
    geraetetyp: Optional["Geraetetyp"] = None
    """Der speziellere Typ eines Gerätes, beispielsweise Stromwandler"""



# pylint:disable=missing-module-docstring

from bo4e.enum.strenum import StrEnum


class Verbrauchsart(StrEnum):
    """
    Verbrauchsart einer Marktlokation.
    """

    KL = "KL"
    """Kraft/Licht"""
    KLW = "KLW"
    """Kraft/Licht/Wärme"""
    KLWS = "KLWS"
    """Kraft/Licht/Wärme/Speicherheizung"""
    W = "W"
    """Wärme"""
    WS = "WS"
    """Wärme/Speicherheizung"""



"""
Contains Marktlokation class
"""

# pylint: disable=too-many-instance-attributes, too-few-public-methods
from typing import TYPE_CHECKING, Annotated, Literal, Optional

from pydantic import Field

from ..enum.botyp import BoTyp
from ..utils import postprocess_docstring
from .geschaeftsobjekt import Geschaeftsobjekt

if TYPE_CHECKING:
    from ..com.adresse import Adresse
    from ..com.geokoordinaten import Geokoordinaten
    from ..com.katasteradresse import Katasteradresse
    from ..com.verbrauch import Verbrauch
    from ..com.zaehlwerk import Zaehlwerk
    from ..enum.bilanzierungsmethode import Bilanzierungsmethode
    from ..enum.energierichtung import Energierichtung
    from ..enum.gasqualitaet import Gasqualitaet
    from ..enum.gebiettyp import Gebiettyp
    from ..enum.kundentyp import Kundentyp
    from ..enum.netzebene import Netzebene
    from ..enum.sparte import Sparte
    from ..enum.verbrauchsart import Verbrauchsart
    from .geschaeftspartner import Geschaeftspartner
    from .lokationszuordnung import Lokationszuordnung

# pylint: disable=no-name-in-module


@postprocess_docstring
class Marktlokation(Geschaeftsobjekt):
    """
    Object containing information about a Marktlokation

    .. raw:: html

        <object data="../_static/images/bo4e/bo/Marktlokation.svg" type="image/svg+xml"></object>

    .. HINT::
        `Marktlokation JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/BO4E/BO4E-Schemas/{__gh_version__}/src/bo4e_schemas/bo/Marktlokation.json>`_

    """

    typ: Annotated[Literal[BoTyp.MARKTLOKATION], Field(alias="_typ")] = BoTyp.MARKTLOKATION

    marktlokations_id: Optional[str] = None
    """Identifikationsnummer einer Marktlokation, an der Energie entweder verbraucht, oder erzeugt wird."""
    sparte: Optional["Sparte"] = None
    """Sparte der Marktlokation, z.B. Gas oder Strom"""
    energierichtung: Optional["Energierichtung"] = None
    """Kennzeichnung, ob Energie eingespeist oder entnommen (ausgespeist) wird"""
    bilanzierungsmethode: Optional["Bilanzierungsmethode"] = None
    """Die Bilanzierungsmethode, RLM oder SLP"""
    netzebene: Optional["Netzebene"] = None
    """
    Netzebene, in der der Bezug der Energie erfolgt.
    Bei Strom Spannungsebene der Lieferung, bei Gas Druckstufe.
    Beispiel Strom: Niederspannung Beispiel Gas: Niederdruck.
    """

    verbrauchsart: Optional["Verbrauchsart"] = None
    """Verbrauchsart der Marktlokation."""
    ist_unterbrechbar: Optional[bool] = None
    """Gibt an, ob es sich um eine unterbrechbare Belieferung handelt"""
    netzbetreibercodenr: Optional[str] = None
    """Codenummer des Netzbetreibers, an dessen Netz diese Marktlokation angeschlossen ist."""
    gebietstyp: Optional["Gebiettyp"] = None
    """Typ des Netzgebietes, z.B. Verteilnetz"""
    netzgebietsnr: Optional[str] = None
    """Die ID des Gebietes in der ene't-Datenbank"""  # todo: rename to "id" (see 2021-12-15 update)
    bilanzierungsgebiet: Optional[str] = None
    """Bilanzierungsgebiet, dem das Netzgebiet zugeordnet ist - im Falle eines Strom Netzes"""
    grundversorgercodenr: Optional[str] = None
    """Codenummer des Grundversorgers, der für diese Marktlokation zuständig ist"""
    gasqualitaet: Optional["Gasqualitaet"] = None
    """Die Gasqualität in diesem Netzgebiet. H-Gas oder L-Gas. Im Falle eines Gas-Netzes"""
    endkunde: Optional["Geschaeftspartner"] = None
    """Geschäftspartner, dem diese Marktlokation gehört"""

    # only one of the following three optional attributes can be set
    lokationsadresse: Optional["Adresse"] = None
    """Die Adresse, an der die Energie-Lieferung oder -Einspeisung erfolgt"""
    geoadresse: Optional["Geokoordinaten"] = None
    """
    Alternativ zu einer postalischen Adresse kann hier ein Ort mittels Geokoordinaten angegeben werden
    (z.B. zur Identifikation von Sendemasten).
    """
    katasterinformation: Optional["Katasteradresse"] = None
    """
    Alternativ zu einer postalischen Adresse und Geokoordinaten kann hier eine Ortsangabe mittels Gemarkung und
    Flurstück erfolgen.
    """

    kundengruppen: Optional[list["Kundentyp"]] = None
    """Kundengruppen der Marktlokation"""

    regelzone: Optional[str] = None
    """für Strom. Code vom EIC, https://www.entsoe.eu/data/energy-identification-codes-eic/eic-approved-codes/"""
    marktgebiet: Optional[str] = None
    """für Gas. Code vom EIC, https://www.entsog.eu/data/data-portal/codes-list"""
    zaehlwerke: Optional[list["Zaehlwerk"]] = None
    verbrauchsmengen: Optional[list["Verbrauch"]] = None
    zaehlwerke_der_beteiligten_marktrolle: Optional[list["Zaehlwerk"]] = None

    lokationszuordnungen: Optional[list["Lokationszuordnung"]] = None
    """Lokationszuordnung, um bspw. die zugehörigen Messlokationen anzugeben"""
    lokationsbuendel_objektcode: Optional[str] = None
    """Lokationsbuendel Code, der die Funktion dieses BOs an der Lokationsbuendelstruktur beschreibt."""



# pylint: disable=missing-module-docstring
from bo4e.enum.strenum import StrEnum


class Ausschreibungstyp(StrEnum):
    """
    Aufzählung für die Typisierung von Ausschreibungen.
    """

    PRIVATRECHTLICH = "PRIVATRECHTLICH"
    """privat-rechtlich"""
    OEFFENTLICHRECHTLICH = "OEFFENTLICHRECHTLICH"
    """öffentlich-rechtlich"""
    EUROPAWEIT = "EUROPAWEIT"
    """europaweit"""



"""
Contains Zeitreihenwert class
"""

from decimal import Decimal
from typing import TYPE_CHECKING, Annotated, Literal, Optional

from pydantic import Field

from ..enum.comtyp import ComTyp
from ..utils import postprocess_docstring
from .com import COM

if TYPE_CHECKING:
    from ..com.zeitraum import Zeitraum
    from ..enum.messwertstatus import Messwertstatus
    from ..enum.messwertstatuszusatz import Messwertstatuszusatz


# pylint: disable=too-few-public-methods
@postprocess_docstring
class Zeitreihenwert(COM):
    """
    Abbildung eines Zeitreihenwertes bestehend aus Zeitraum, Wert und Statusinformationen.

    .. raw:: html

        <object data="../_static/images/bo4e/com/Zeitreihenwert.svg" type="image/svg+xml"></object>

    .. HINT::
        `Zeitreihenwert JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/BO4E/BO4E-Schemas/{__gh_version__}/src/bo4e_schemas/com/Zeitreihenwert.json>`_

    """

    typ: Annotated[Literal[ComTyp.ZEITREIHENWERT], Field(alias="_typ")] = ComTyp.ZEITREIHENWERT

    zeitraum: Optional["Zeitraum"] = None
    """Zeitraum für das Messintervall"""
    wert: Optional[Decimal] = None
    """Der in dem Zeitraum gültige Wert."""
    status: Optional["Messwertstatus"] = None
    """Der Status gibt an, wie der Wert zu interpretieren ist, z.B. in Berechnungen."""
    statuszusatz: Optional["Messwertstatuszusatz"] = None
    """Eine Zusatzinformation zum Status, beispielsweise ein Grund für einen fehlenden Wert."""



import pytest

from bo4e import Betrag, Fremdkosten, Fremdkostenblock, Zeitraum
from tests.serialization_helper import assert_serialization_roundtrip


class TestFremdkosten:
    @pytest.mark.parametrize(
        "fremdkosten",
        [
            pytest.param(
                Fremdkosten(
                    gueltigkeit=Zeitraum(),
                    summe_kosten=Betrag(),
                    kostenbloecke=[Fremdkostenblock()],
                ),
                id="all attributes at first level",
            ),
        ],
    )
    def test_serialization_roundtrip(self, fremdkosten: Fremdkosten) -> None:
        """
        Test de-/serialisation
        """
        assert_serialization_roundtrip(fremdkosten)



# pylint: disable=missing-module-docstring
from bo4e.enum.strenum import StrEnum


class Oekozertifikat(StrEnum):
    """
    Zertifikate für Ökostrom von verschiedenen Herausgebern.
    """

    CMS_EE01 = "CMS_EE01"
    """CMS_EE01"""
    CMS_EE02 = "CMS_EE02"
    """CMS_EE02"""
    EECS = "EECS"
    """EECS"""
    FRAUNHOFER = "FRAUNHOFER"
    """FRAUNHOFER"""
    BET = "BET"
    """BET"""
    KLIMA_INVEST = "KLIMA_INVEST"
    """KLIMA_INVEST"""
    LGA = "LGA"
    """LGA"""
    FREIBERG = "FREIBERG"
    """FREIBERG"""
    RECS = "RECS"
    """RECS"""
    REGS_EGL = "REGS_EGL"
    """REGS_EGL"""
    TUEV = "TUEV"
    """TUEV"""
    TUEV_HESSEN = "TUEV_HESSEN"
    """TUEV_HESSEN"""
    TUEV_NORD = "TUEV_NORD"
    """TUEV_NORD"""
    TUEV_RHEINLAND = "TUEV_RHEINLAND"
    """TUEV_RHEINLAND"""
    TUEV_SUED = "TUEV_SUED"
    """TUEV_SUED"""
    TUEV_SUED_EE01 = "TUEV_SUED_EE01"
    """TUEV_SUED_EE01"""
    TUEV_SUED_EE02 = "TUEV_SUED_EE02"
    """TUEV_SUED_EE02"""



import pytest

from bo4e import AufAbschlagProOrt, AufAbschlagstaffelProOrt
from tests.serialization_helper import assert_serialization_roundtrip


class TestAufAbschlagProOrt:
    @pytest.mark.parametrize(
        "aufabschlagproort",
        [
            pytest.param(
                AufAbschlagProOrt(
                    postleitzahl="01187",
                    ort="Dresden",
                    netznr="2",
                    staffeln=[AufAbschlagstaffelProOrt()],
                ),
            )
        ],
    )
    def test_serialization_roundtrip(
        self,
        aufabschlagproort: AufAbschlagProOrt,
    ) -> None:
        """
        Test de-/serialisation of AufAbschlagProOrt.
        """
        assert_serialization_roundtrip(aufabschlagproort)



from datetime import datetime, time

import pytest

from bo4e import (
    Gueltigkeitstyp,
    KriteriumWert,
    Preisgarantietyp,
    RegionaleGueltigkeit,
    RegionalePreisgarantie,
    Tarifregionskriterium,
    Zeitraum,
)
from tests.serialization_helper import assert_serialization_roundtrip


class TestRegionalePreisgarantie:
    @pytest.mark.parametrize(
        "regionale_preisgarantie",
        [
            pytest.param(
                RegionalePreisgarantie(
                    preisgarantietyp=Preisgarantietyp.NUR_ENERGIEPREIS,
                    zeitliche_gueltigkeit=Zeitraum(
                        startdatum=datetime(2011, 2, 5),
                        startuhrzeit=time(16, 43),
                        enddatum=datetime(2021, 7, 30),
                        enduhrzeit=time(16, 43),
                    ),
                    regionale_gueltigkeit=RegionaleGueltigkeit(
                        gueltigkeitstyp=Gueltigkeitstyp.NUR_IN,
                        kriteriums_werte=[KriteriumWert(kriterium=Tarifregionskriterium.POSTLEITZAHL, wert="01069")],
                    ),
                ),
            ),
        ],
    )
    def test_regionale_preisgarantie_serialization_roundtrip(
        self, regionale_preisgarantie: RegionalePreisgarantie
    ) -> None:
        """
        Test de-/serialisation of RegionalePreisgarantie with minimal attributes.
        """
        assert_serialization_roundtrip(regionale_preisgarantie)



# pylint: disable=missing-module-docstring
from bo4e.enum.strenum import StrEnum


class Tarifregionskriterium(StrEnum):
    """
    Mit diesen Kriterien können regionale Bereiche definiert werden.
    """

    NETZ_NUMMER = "NETZ_NUMMER"
    """Netznummer"""
    POSTLEITZAHL = "POSTLEITZAHL"
    """Postleitzahl"""
    ORT = "ORT"
    """Ort"""
    GRUNDVERSORGER_NUMMER = "GRUNDVERSORGER_NUMMER"
    """Nummer des Grundversorgers"""
    REGION = "REGION"
    """Referenz auf ein BO Region (URL)"""



# pylint: disable=missing-module-docstring
from bo4e.enum.strenum import StrEnum


class Mengenoperator(StrEnum):
    """
    Angabe, wie eine Menge in Bezug auf einen Wert zu bilden ist.
    """

    KLEINER_ALS = "KLEINER_ALS"
    """KLEINER_ALS"""
    GROESSER_ALS = "GROESSER_ALS"
    """GROESSER_ALS"""
    GLEICH = "GLEICH"
    """GLEICH"""
